import math
from unittest.mock import MagicMock

import pytest

from opentrons.legacy_commands import types
from opentrons.protocol_api import InstrumentContext
from opentrons.protocol_api.core.legacy.deck import Deck
from opentrons.protocols.duration.estimator import (
    TEMP_MOD_HIGH_THRESH,
    TEMP_MOD_LOW_THRESH,
    TEMP_MOD_RATE_HIGH_AND_ABOVE,
    TEMP_MOD_RATE_LOW_TO_HIGH,
    TEMP_MOD_RATE_ZERO_TO_LOW,
    THERMO_HIGH_THRESH,
    THERMO_LOW_THRESH,
    DurationEstimator,
)
from opentrons.types import Location, Point

messageDataBefore: types.CommandMessageFields = {
    "$": "before",
    "id": "b",
    "error": None,
}

messageDataAfter: types.CommandMessageFields = {"$": "after", "id": "a", "error": None}


@pytest.fixture
def subject() -> DurationEstimator:
    """The test subject."""
    return DurationEstimator()


@pytest.fixture
def mock_instrument() -> MagicMock:
    """A mock instrument context."""
    return MagicMock(spec=InstrumentContext)


@pytest.fixture
def mock_location() -> MagicMock:
    """A mock well."""
    return MagicMock(spec=Location)


def test_ignore_before(subject: DurationEstimator) -> None:
    """It should ignore before commands."""
    message = types.DelayMessage(
        payload=types.DelayCommandPayload(minutes=1, seconds=1, text=""),
        name=types.DELAY,
        **messageDataBefore,
    )
    subject.on_message(message)
    assert subject.get_total_duration() == 0


def test_pick_up_tip(
    subject: DurationEstimator, mock_instrument: MagicMock, mock_location: MagicMock
) -> None:
    """It should time a pick up tip."""
    mock_instrument.default_speed = 1
    subject._last_deckslot = "1"
    mock_location.labware.first_parent.return_value = "1"
    message = types.PickUpTipMessage(
        payload=types.PickUpTipCommandPayload(
            location=mock_location, instrument=mock_instrument, text=""
        ),
        name=types.PICK_UP_TIP,
        **messageDataAfter,
    )
    subject.on_message(message)
    assert subject.get_total_duration() == 4.5


def test_drop_tip(
    subject: DurationEstimator, mock_instrument: MagicMock, mock_location: MagicMock
) -> None:
    """It should time a drop tip."""
    mock_instrument.default_speed = 1
    # Movement in same slot
    subject._last_deckslot = "1"
    mock_location.labware.first_parent.return_value = "1"
    message = types.DropTipMessage(
        payload=types.DropTipCommandPayload(
            location=mock_location, instrument=mock_instrument, text=""
        ),
        name=types.DROP_TIP,
        **messageDataAfter,
    )
    subject.on_message(message)
    assert subject.get_total_duration() == 10.5


def test_blow_out(subject: DurationEstimator, mock_instrument: MagicMock) -> None:
    """It should time a blowout."""
    message = types.BlowOutMessage(
        payload=types.BlowOutCommandPayload(
            location=None, instrument=mock_instrument, text=""
        ),
        name=types.BLOW_OUT,
        **messageDataAfter,
    )
    subject.on_message(message)
    assert subject.get_total_duration() == 0.5


def test_touch_tip(subject: DurationEstimator, mock_instrument: MagicMock) -> None:
    """It should time a touch tip."""
    message = types.TouchTipMessage(
        payload=types.TouchTipCommandPayload(instrument=mock_instrument, text=""),
        name=types.TOUCH_TIP,
        **messageDataAfter,
    )
    subject.on_message(message)
    assert subject.get_total_duration() == 0.5


def test_delay(subject: DurationEstimator) -> None:
    """It should time a delay."""
    message = types.DelayMessage(
        payload=types.DelayCommandPayload(minutes=1, seconds=1, text=""),
        name=types.DELAY,
        **messageDataAfter,
    )
    subject.on_message(message)
    assert subject.get_total_duration() == 61


@pytest.fixture
def mock_deck() -> MagicMock:
    """A mock deck fixture"""

    def position_for(slot: str) -> Location:
        # zero based
        s = int(slot) - 1
        row = int(s / 3)
        col = int(s % 3)
        return Location(point=Point(x=col * 10, y=row * 100), labware=None)

    m = MagicMock(spec=Deck)
    m.position_for.side_effect = position_for
    return m


@pytest.mark.parametrize(
    argnames=["current_slot", "previous_slot", "speed", "expected_duration"],
    argvalues=[
        # Within same slot
        ["12", "12", 1, 0.5],
        # Same row sqrt(10**2) / 2
        ["1", "2", 2, 5],
        # Same col sqrt(100**2) / 2
        ["1", "4", 2, 50],
        # Extremes
        ["1", "12", 2, math.sqrt(20**2 + 300**2) / 2],
        # Same row sqrt(10**2) / 2
        ["3", "2", 2, 5],
        # Same col sqrt(100**2) / 2
        ["5", "2", 2, 50],
        # Extremes
        ["12", "1", 2, math.sqrt(20**2 + 300**2) / 2],
    ],
)
def test_calc_deck_movement_time(
    subject: DurationEstimator,
    mock_deck: MagicMock,
    current_slot: str,
    previous_slot: str,
    speed: float,
    expected_duration: float,
) -> None:
    """It should calculate deck movement correctly."""
    assert (
        subject.calc_deck_movement_time(mock_deck, current_slot, previous_slot, speed)
        == expected_duration
    )


@pytest.mark.parametrize(
    argnames=["current", "target", "expected_duration"],
    argvalues=[
        # No change in temperature
        [1, 1, 0],
        # To a high temperature from a high temperature
        [
            TEMP_MOD_HIGH_THRESH,
            TEMP_MOD_HIGH_THRESH + 1,
            1 / TEMP_MOD_RATE_HIGH_AND_ABOVE,
        ],
        # To a medium temperature from a medium temperature
        [
            TEMP_MOD_LOW_THRESH + 1,
            TEMP_MOD_HIGH_THRESH - 1,
            ((TEMP_MOD_HIGH_THRESH - 1) - (TEMP_MOD_LOW_THRESH + 1))
            / TEMP_MOD_RATE_LOW_TO_HIGH,
        ],
        # To a low temperature from a low temperature
        [
            TEMP_MOD_LOW_THRESH - 1,
            TEMP_MOD_LOW_THRESH - 2,
            1 / TEMP_MOD_RATE_ZERO_TO_LOW,
        ],
    ],
)
def test_temperature_module(
    subject: DurationEstimator, current: float, target: float, expected_duration: float
) -> None:
    """It should compute the duration of a temperature change correctly."""
    assert subject.temperature_module(current, target) == expected_duration


def test_thermocycler_set_lid_temp(subject: DurationEstimator) -> None:
    """It should compute the duration of a set lid temp."""
    message = types.ThermocyclerSetLidTempMessage(
        payload=types.ThermocyclerSetLidTempCommandPayload(text=""),
        name=types.THERMOCYCLER_SET_LID_TEMP,
        **messageDataAfter,
    )
    subject.on_message(message)
    assert subject.get_total_duration() == 60


def test_thermocycler_lid_close(subject: DurationEstimator) -> None:
    """It should compute the duration of a lid close."""
    message = types.ThermocyclerCloseMessage(
        payload=types.ThermocyclerCloseCommandPayload(text=""),
        name=types.THERMOCYCLER_CLOSE,
        **messageDataAfter,
    )
    subject.on_message(message)
    assert subject.get_total_duration() == 24


def test_thermocycler_lid_open(subject: DurationEstimator) -> None:
    """It should compute the duration of a lid open."""
    message = types.ThermocyclerOpenMessage(
        payload=types.ThermocyclerOpenCommandPayload(text=""),
        name=types.THERMOCYCLER_OPEN,
        **messageDataAfter,
    )
    subject.on_message(message)
    assert subject.get_total_duration() == 24


def test_thermocycler_deactivate_lid(subject: DurationEstimator) -> None:
    """It should compute the duration of a lid open."""
    message = types.ThermocyclerDeactivateLidMessage(
        payload=types.ThermocyclerDeactivateLidCommandPayload(text=""),
        name=types.THERMOCYCLER_DEACTIVATE_LID,
        **messageDataAfter,
    )
    subject.on_message(message)
    assert subject.get_total_duration() == 23


@pytest.mark.parametrize(
    argnames=["current", "target", "expected_duration"],
    argvalues=[
        # No change in temperature
        [1, 1, 0],
        # Heating above high threshold
        [10, THERMO_HIGH_THRESH + 5, (5 / 2) + ((THERMO_HIGH_THRESH - 10) / 4)],
        # Heating below high threshold
        [THERMO_HIGH_THRESH - 10, THERMO_HIGH_THRESH - 5, (5 / 4)],
        # Cooling above high threshold
        [THERMO_HIGH_THRESH + 20, THERMO_HIGH_THRESH + 10, (10 / 2)],
        # Cooling above low threshold
        [THERMO_LOW_THRESH + 20, THERMO_LOW_THRESH + 10, 10],
        # Cooling below low threshold
        [THERMO_LOW_THRESH - 10, THERMO_LOW_THRESH - 20, (10 / 0.5) + (20 / 0.1)],
    ],
)
def test_thermocycler_handler(
    subject: DurationEstimator, current: float, target: float, expected_duration: float
) -> None:
    """It should compute the duration of a temperature change correctly."""
    assert subject.thermocycler_handler(current, target) == expected_duration



"""Tests for TaskState+TaskStore+TaskView trifecta."""

import asyncio
from datetime import datetime

import pytest

from opentrons.protocol_engine.actions.actions import (
    FinishTaskAction,
    StartTaskAction,
)
from opentrons.protocol_engine.errors import ErrorOccurrence
from opentrons.protocol_engine.state.tasks import TaskStore, TaskView
from opentrons.protocol_engine.types import FinishedTask, Task


@pytest.fixture
def subject() -> TaskStore:
    """Create a TaskStore fixture."""
    return TaskStore()


async def test_get_current(subject: TaskStore) -> None:
    """It should get a current task by ID."""
    task_id = "task-123"
    timestamp = datetime.now()
    asyncio_task = asyncio.create_task(asyncio.sleep(0))
    subject._state.current_tasks_by_id[task_id] = Task(
        id=task_id,
        createdAt=timestamp,
        asyncioTask=asyncio_task,
    )

    view = TaskView(subject._state)
    result = view.get_current(task_id)

    assert isinstance(result, Task)
    assert result.id == task_id
    assert result.createdAt == timestamp
    assert result.asyncioTask is asyncio_task
    await asyncio_task


async def test_get_finished(subject: TaskStore) -> None:
    """It should get a finished task by ID."""
    task_id = "task-123"
    timestamp = datetime.now()
    timestamp2 = datetime.now()
    error = None
    asyncio_task = asyncio.create_task(asyncio.sleep(0))

    subject._state.finished_tasks_by_id[task_id] = FinishedTask(
        id=task_id, createdAt=timestamp, finishedAt=timestamp2, error=error
    )

    view = TaskView(subject._state)
    result = view.get_finished(task_id)

    assert isinstance(result, FinishedTask)
    assert result.id == task_id
    assert result.createdAt == timestamp
    assert result.finishedAt == timestamp2
    assert result.error == error
    await asyncio_task


async def test_get(subject: TaskStore) -> None:
    """It should get all tasks (finished and current)."""
    task_id_current = "task-current"
    timestamp = datetime.now()
    asyncio_task_1 = asyncio.create_task(asyncio.sleep(0))
    subject._state.current_tasks_by_id[task_id_current] = Task(
        id=task_id_current,
        createdAt=timestamp,
        asyncioTask=asyncio_task_1,
    )

    task_id_finished = "task-finished"
    timestamp2 = datetime.now()
    timestamp3 = datetime.now()
    error = None
    subject._state.finished_tasks_by_id[task_id_finished] = FinishedTask(
        id=task_id_finished, createdAt=timestamp2, finishedAt=timestamp3, error=error
    )
    view = TaskView(subject._state)
    # Check current task
    result = view.get("task-current")
    assert isinstance(result, Task)
    assert result.id == task_id_current
    assert result.createdAt == timestamp
    assert result.asyncioTask is asyncio_task_1
    await asyncio_task_1
    # Check finished task
    result_finished = view.get("task-finished")
    assert isinstance(result_finished, FinishedTask)
    assert result_finished.id == task_id_finished
    assert result_finished.createdAt == timestamp2
    assert result_finished.finishedAt == timestamp3
    assert result_finished.error == error


async def test_get_summary(subject: TaskStore) -> None:
    """It should get a summary of all tasks."""
    task_id_1 = "task-123"
    task_id_2 = "task-456"
    task_id_3 = "task-789"
    timestamp_1 = datetime.now()
    timestamp_2 = datetime.now()
    timestamp_3 = datetime.now()
    timestamp_4 = datetime.now()
    asyncio_task_1 = asyncio.create_task(asyncio.sleep(0))
    asyncio_task_2 = asyncio.create_task(asyncio.sleep(0))

    subject._state.current_tasks_by_id[task_id_1] = Task(
        id=task_id_1,
        createdAt=timestamp_1,
        asyncioTask=asyncio_task_1,
    )
    subject._state.current_tasks_by_id[task_id_2] = Task(
        id=task_id_2,
        createdAt=timestamp_2,
        asyncioTask=asyncio_task_2,
    )
    subject._state.finished_tasks_by_id[task_id_3] = FinishedTask(
        id=task_id_1,
        createdAt=timestamp_3,
        finishedAt=timestamp_4,
        error=None,
    )

    view = TaskView(subject._state)
    summary = view.get_summary()

    assert len(summary) == 3
    assert summary[0].id == task_id_1
    assert summary[0].createdAt == timestamp_1
    assert summary[0].finishedAt is None
    assert summary[0].error is None

    assert summary[1].id == task_id_2
    assert summary[1].createdAt == timestamp_2
    assert summary[1].finishedAt is None
    assert summary[1].error is None

    assert summary[2].id == task_id_3
    assert summary[2].createdAt == timestamp_3
    assert summary[2].finishedAt is timestamp_4
    assert summary[2].error is None

    await asyncio.gather(asyncio_task_1, asyncio_task_2)


async def test_handle_start_task_action(subject: TaskStore) -> None:
    """It should store data about a start task action."""
    timestamp_1 = datetime.now()
    task_id_1 = "task123"
    asyncio_task_1 = asyncio.create_task(asyncio.sleep(0))
    await asyncio.gather(asyncio_task_1)

    action_started = StartTaskAction(
        task=Task(
            id=task_id_1,
            createdAt=timestamp_1,
            asyncioTask=asyncio_task_1,
        )
    )
    subject.handle_action(action_started)
    task = subject._state.current_tasks_by_id[task_id_1]
    assert task.id == task_id_1
    assert task.createdAt == timestamp_1
    assert task.asyncioTask == asyncio_task_1


async def test_handle_finish_task_action(subject: TaskStore) -> None:
    """It should store data about a finish task action."""
    timestamp_1 = datetime.now()
    timestamp_2 = datetime.now()
    task_id_1 = "task123"
    asyncio_task_1 = asyncio.create_task(asyncio.sleep(0))
    await asyncio.gather(asyncio_task_1)

    action_started = StartTaskAction(
        task=Task(
            id=task_id_1,
            createdAt=timestamp_1,
            asyncioTask=asyncio_task_1,
        )
    )
    subject.handle_action(action_started)
    assert task_id_1 in subject._state.current_tasks_by_id
    action_finished = FinishTaskAction(
        task_id=task_id_1, finished_at=timestamp_2, error=None
    )
    subject.handle_action(action_finished)
    task_finished = subject._state.finished_tasks_by_id[task_id_1]
    assert isinstance(task_finished, FinishedTask)
    assert task_finished.id == task_id_1
    assert task_finished.createdAt == timestamp_1
    assert task_finished.finishedAt == timestamp_2
    assert task_finished.error is None
    assert task_id_1 not in subject._state.current_tasks_by_id


async def test_all_tasks_finished_or_any_task_failed(subject: TaskStore) -> None:
    """It should return false if there are no finished tasks and true if there are."""
    # Returns False because only task is current
    task_id_current = "task-current"
    timestamp = datetime.now()
    asyncio_task = asyncio.create_task(asyncio.sleep(0))
    subject._state.current_tasks_by_id[task_id_current] = Task(
        id=task_id_current,
        createdAt=timestamp,
        asyncioTask=asyncio_task,
    )
    view = TaskView(subject._state)
    result = view.all_tasks_finished_or_any_task_failed(task_id_current)
    assert result is False
    await asyncio_task
    # returns true because task is finished
    task_id_finished = "task-finished"
    timestamp2 = datetime.now()
    timestamp3 = datetime.now()
    error = None
    asyncio_task = asyncio.create_task(asyncio.sleep(0))

    subject._state.finished_tasks_by_id[task_id_finished] = FinishedTask(
        id=task_id_finished, createdAt=timestamp2, finishedAt=timestamp3, error=error
    )
    view = TaskView(subject._state)
    result = view.all_tasks_finished_or_any_task_failed({task_id_finished})
    assert result

    # returns true because 1 task has error
    subject._state.finished_tasks_by_id[task_id_finished] = FinishedTask(
        id=task_id_finished, createdAt=timestamp2, finishedAt=timestamp3, error=error
    )
    task_id_error = "task-error"
    timestamp2 = datetime.now()
    timestamp3 = datetime.now()
    error = ErrorOccurrence(
        id="error",
        createdAt=datetime.now(),
        errorType="TaskFailedError",
        detail="detail",
    )
    asyncio_task = asyncio.create_task(asyncio.sleep(0))

    subject._state.finished_tasks_by_id[task_id_error] = FinishedTask(
        id=task_id_error, createdAt=timestamp2, finishedAt=timestamp3, error=error
    )

    view = TaskView(subject._state)
    result = view.all_tasks_finished_or_any_task_failed(
        {task_id_finished, task_id_error}
    )
    assert result


def test_get_failed_tasks(subject: TaskStore) -> None:
    """It should return a list of failed tasks."""
    task_id_error = "task-error"
    timestamp2 = datetime.now()
    timestamp3 = datetime.now()
    error = ErrorOccurrence(
        id="error",
        createdAt=datetime.now(),
        errorType="TaskFailedError",
        detail="detail",
    )
    subject._state.finished_tasks_by_id[task_id_error] = FinishedTask(
        id=task_id_error, createdAt=timestamp2, finishedAt=timestamp3, error=error
    )
    task_id_error_2 = "task-error-2"
    timestamp2 = datetime.now()
    timestamp3 = datetime.now()
    error = ErrorOccurrence(
        id="error",
        createdAt=datetime.now(),
        errorType="TaskFailedError",
        detail="detail",
    )
    subject._state.finished_tasks_by_id[task_id_error_2] = FinishedTask(
        id=task_id_error_2, createdAt=timestamp2, finishedAt=timestamp3, error=error
    )
    view = TaskView(subject._state)
    result = view.get_failed_tasks({task_id_error})

    assert len(result) == 1
    assert result[0] == task_id_error



"""Tests for the ProtocolEngine class."""

import inspect
from datetime import datetime
from typing import Any, cast
from unittest.mock import sentinel

import pytest
from decoy import Decoy

from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from opentrons_shared_data.robot.types import RobotType

from opentrons.hardware_control import HardwareControlAPI, OT2HardwareControlAPI
from opentrons.hardware_control.modules import MagDeck, TempDeck
from opentrons.hardware_control.types import PauseType as HardwarePauseType
from opentrons.protocol_engine import (
    ProtocolEngine,
    commands,
    labware_offset_standardization,
    slot_standardization,
)
from opentrons.protocol_engine.actions import (
    ActionDispatcher,
    AddAddressableAreaAction,
    AddCameraSettingsAction,
    AddLabwareDefinitionAction,
    AddLabwareOffsetAction,
    AddLiquidAction,
    AddModuleAction,
    FinishAction,
    FinishErrorDetails,
    HardwareStoppedAction,
    PauseAction,
    PauseSource,
    PlayAction,
    QueueCommandAction,
    ResumeFromRecoveryAction,
    StopAction,
)
from opentrons.protocol_engine.actions.actions import SetErrorRecoveryPolicyAction
from opentrons.protocol_engine.errors import ErrorOccurrence, ProtocolCommandFailedError
from opentrons.protocol_engine.errors.exceptions import (
    CommandNotAllowedError,
)
from opentrons.protocol_engine.execution import (
    DoorWatcher,
    HardwareStopper,
    QueueWorker,
)
from opentrons.protocol_engine.plugins import AbstractPlugin, PluginStarter
from opentrons.protocol_engine.resources import (
    CameraProvider,
    FileProvider,
    ModelUtils,
    ModuleDataProvider,
)
from opentrons.protocol_engine.resources.camera_provider import CameraSettings
from opentrons.protocol_engine.state.config import Config
from opentrons.protocol_engine.state.state import StateStore
from opentrons.protocol_engine.state.update_types import StateUpdate
from opentrons.protocol_engine.types import (
    DeckType,
    LabwareOffset,
    LabwareOffsetCreate,
    LabwareOffsetCreateInternal,
    LabwareOffsetVector,
    LabwareUri,
    LegacyLabwareOffsetCreate,
    LegacyLabwareOffsetLocation,
    Liquid,
    ModuleDefinition,
    ModuleModel,
    OnAddressableAreaOffsetLocationSequenceComponent,
    PostRunHardwareState,
)
from opentrons.types import DeckSlotName


@pytest.fixture
def state_store(decoy: Decoy) -> StateStore:
    """Get a mock StateStore."""
    return decoy.mock(cls=StateStore)


@pytest.fixture
def action_dispatcher(decoy: Decoy) -> ActionDispatcher:
    """Get a mock ActionDispatcher."""
    return decoy.mock(cls=ActionDispatcher)


@pytest.fixture
def plugin_starter(decoy: Decoy) -> PluginStarter:
    """Get a mock PluginStarter."""
    return decoy.mock(cls=PluginStarter)


@pytest.fixture
def queue_worker(decoy: Decoy) -> QueueWorker:
    """Get a mock QueueWorker."""
    return decoy.mock(cls=QueueWorker)


@pytest.fixture
def model_utils(decoy: Decoy) -> ModelUtils:
    """Get mock ModelUtils."""
    return decoy.mock(cls=ModelUtils)


@pytest.fixture
def hardware_api(decoy: Decoy) -> HardwareControlAPI:
    """Get a mock HardwareControlAPI."""
    return decoy.mock(cls=OT2HardwareControlAPI)


@pytest.fixture
def hardware_stopper(decoy: Decoy) -> HardwareStopper:
    """Get a mock HardwareStopper."""
    return decoy.mock(cls=HardwareStopper)


@pytest.fixture
def door_watcher(decoy: Decoy) -> DoorWatcher:
    """Get a mock DoorWatcher."""
    return decoy.mock(cls=DoorWatcher)


@pytest.fixture
def module_data_provider(decoy: Decoy) -> ModuleDataProvider:
    """Get a mock ModuleDataProvider."""
    return decoy.mock(cls=ModuleDataProvider)


@pytest.fixture
def file_provider(decoy: Decoy) -> FileProvider:
    """Get a mock FileProvider."""
    return decoy.mock(cls=FileProvider)


@pytest.fixture
def camera_provider(decoy: Decoy) -> CameraProvider:
    """Get a mock CameraProvider."""
    return decoy.mock(cls=CameraProvider)


@pytest.fixture(autouse=True)
def _mock_slot_standardization_module(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out opentrons.protocol_engine.slot_standardization functions."""
    for name, func in inspect.getmembers(slot_standardization, inspect.isfunction):
        monkeypatch.setattr(slot_standardization, name, decoy.mock(func=func))


@pytest.fixture(autouse=True)
def _mock_labware_offset_standardization_module(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out opentrons.labware_offset_standardization functions."""
    for name, func in inspect.getmembers(
        labware_offset_standardization, inspect.isfunction
    ):
        monkeypatch.setattr(labware_offset_standardization, name, decoy.mock(func=func))


@pytest.fixture(autouse=True)
def _mock_hash_command_params_module(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    hash_command_params = commands.hash_protocol_command_params
    monkeypatch.setattr(
        commands, "hash_protocol_command_params", decoy.mock(func=hash_command_params)
    )


@pytest.fixture
def subject(
    hardware_api: HardwareControlAPI,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    plugin_starter: PluginStarter,
    queue_worker: QueueWorker,
    model_utils: ModelUtils,
    hardware_stopper: HardwareStopper,
    door_watcher: DoorWatcher,
    module_data_provider: ModuleDataProvider,
    file_provider: FileProvider,
    camera_provider: CameraProvider,
) -> ProtocolEngine:
    """Get a ProtocolEngine test subject with its dependencies stubbed out."""
    return ProtocolEngine(
        hardware_api=hardware_api,
        state_store=state_store,
        action_dispatcher=action_dispatcher,
        plugin_starter=plugin_starter,
        queue_worker=queue_worker,
        model_utils=model_utils,
        hardware_stopper=hardware_stopper,
        door_watcher=door_watcher,
        module_data_provider=module_data_provider,
        file_provider=file_provider,
        camera_provider=camera_provider,
    )


def test_create_starts_background_tasks(
    decoy: Decoy,
    queue_worker: QueueWorker,
    door_watcher: DoorWatcher,
    subject: ProtocolEngine,
) -> None:
    """It should start the queue worker upon creation."""
    decoy.verify(queue_worker.start(), door_watcher.start())


def test_add_command(
    decoy: Decoy,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
    subject: ProtocolEngine,
) -> None:
    """It should add a command to the state from a request."""
    created_at = datetime(year=2021, month=1, day=1)
    original_request = commands.WaitForResumeCreate(
        params=commands.WaitForResumeParams()
    )
    standardized_request = commands.HomeCreate(
        params=commands.HomeParams(), intent=commands.CommandIntent.PROTOCOL
    )
    queued = commands.Home(
        id="command-id",
        key="command-key",
        status=commands.CommandStatus.QUEUED,
        createdAt=created_at,
        params=commands.HomeParams(),
    )

    robot_type: RobotType = "OT-3 Standard"
    decoy.when(state_store.config).then_return(
        Config(robot_type=robot_type, deck_type=DeckType.OT3_STANDARD)
    )

    decoy.when(
        slot_standardization.standardize_command(original_request, robot_type)
    ).then_return(standardized_request)

    decoy.when(model_utils.generate_id()).then_return("command-id")
    decoy.when(model_utils.get_timestamp()).then_return(created_at)
    decoy.when(state_store.commands.get_latest_protocol_command_hash()).then_return(
        "abc"
    )
    decoy.when(
        commands.hash_protocol_command_params(
            create=standardized_request, last_hash="abc"
        )
    ).then_return("123")

    def _stub_queued(*_a: object, **_k: object) -> None:
        decoy.when(state_store.commands.get("command-id")).then_return(queued)

    decoy.when(
        state_store.commands.validate_action_allowed(
            QueueCommandAction(
                command_id="command-id",
                created_at=created_at,
                request=standardized_request,
                request_hash="123",
            )
        )
    ).then_return(
        QueueCommandAction(
            command_id="command-id-validated",
            created_at=created_at,
            request=standardized_request,
            request_hash="456",
        )
    )

    decoy.when(
        action_dispatcher.dispatch(  # type: ignore[func-returns-value]
            QueueCommandAction(
                command_id="command-id-validated",
                created_at=created_at,
                request=standardized_request,
                request_hash="456",
            )
        ),
    ).then_do(_stub_queued)

    result = subject.add_command(original_request)

    assert result == queued


def test_add_fixit_command(
    decoy: Decoy,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
    subject: ProtocolEngine,
) -> None:
    """It should add a fixit command to the state from a request."""
    created_at = datetime(year=2021, month=1, day=1)
    original_request = commands.WaitForResumeCreate(
        params=commands.WaitForResumeParams()
    )
    standardized_request = commands.HomeCreate(
        params=commands.HomeParams(), intent=commands.CommandIntent.FIXIT
    )
    queued = commands.Home(
        id="command-id",
        key="command-key",
        status=commands.CommandStatus.QUEUED,
        createdAt=created_at,
        params=commands.HomeParams(),
    )

    robot_type: RobotType = "OT-3 Standard"
    decoy.when(state_store.config).then_return(
        Config(robot_type=robot_type, deck_type=DeckType.OT3_STANDARD)
    )

    decoy.when(
        slot_standardization.standardize_command(original_request, robot_type)
    ).then_return(standardized_request)

    decoy.when(model_utils.generate_id()).then_return("command-id")
    decoy.when(model_utils.get_timestamp()).then_return(created_at)

    def _stub_queued(*_a: object, **_k: object) -> None:
        decoy.when(state_store.commands.get("command-id")).then_return(queued)

    decoy.when(
        state_store.commands.validate_action_allowed(
            QueueCommandAction(
                command_id="command-id",
                created_at=created_at,
                request=standardized_request,
                request_hash=None,
            )
        )
    ).then_return(
        QueueCommandAction(
            command_id="command-id-validated",
            created_at=created_at,
            request=standardized_request,
            request_hash=None,
        )
    )

    decoy.when(
        action_dispatcher.dispatch(  # type: ignore[func-returns-value]
            QueueCommandAction(
                command_id="command-id-validated",
                created_at=created_at,
                request=standardized_request,
                request_hash=None,
            )
        ),
    ).then_do(_stub_queued)

    result = subject.add_command(original_request)
    assert result == queued


def test_add_fixit_command_raises(
    decoy: Decoy,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
    subject: ProtocolEngine,
) -> None:
    """It should raise if a failedCommandId is supplied without a  fixit command."""
    original_request = commands.WaitForResumeCreate(
        params=commands.WaitForResumeParams()
    )
    standardized_request = commands.HomeCreate(
        params=commands.HomeParams(), intent=commands.CommandIntent.PROTOCOL
    )

    robot_type: RobotType = "OT-3 Standard"
    decoy.when(state_store.config).then_return(
        Config(robot_type=robot_type, deck_type=DeckType.OT3_STANDARD)
    )

    decoy.when(
        slot_standardization.standardize_command(original_request, robot_type)
    ).then_return(standardized_request)

    with pytest.raises(CommandNotAllowedError):
        subject.add_command(original_request, "id-123")


async def test_add_and_execute_command(
    decoy: Decoy,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
    subject: ProtocolEngine,
) -> None:
    """It should add and execute a command from a request."""
    created_at = datetime(year=2021, month=1, day=1)
    original_request = commands.WaitForResumeCreate(
        params=commands.WaitForResumeParams()
    )
    standardized_request = commands.HomeCreate(params=commands.HomeParams())
    queued = commands.Home(
        id="command-id",
        key="command-key",
        status=commands.CommandStatus.QUEUED,
        createdAt=created_at,
        params=commands.HomeParams(),
    )
    completed = commands.Home(
        id="command-id",
        key="command-key",
        status=commands.CommandStatus.SUCCEEDED,
        createdAt=created_at,
        params=commands.HomeParams(),
    )

    robot_type: RobotType = "OT-3 Standard"
    decoy.when(state_store.config).then_return(
        Config(robot_type=robot_type, deck_type=DeckType.OT3_STANDARD)
    )

    decoy.when(
        slot_standardization.standardize_command(original_request, robot_type)
    ).then_return(standardized_request)

    decoy.when(model_utils.generate_id()).then_return("command-id")
    decoy.when(model_utils.get_timestamp()).then_return(created_at)

    def _stub_queued(*_a: object, **_k: object) -> None:
        decoy.when(state_store.commands.get("command-id")).then_return(queued)

    def _stub_completed(*_a: object, **_k: object) -> bool:
        decoy.when(state_store.commands.get("command-id")).then_return(completed)
        return True

    decoy.when(
        state_store.commands.validate_action_allowed(
            QueueCommandAction(
                command_id="command-id",
                created_at=created_at,
                request=standardized_request,
                request_hash=None,
            )
        )
    ).then_return(
        QueueCommandAction(
            command_id="command-id-validated",
            created_at=created_at,
            request=standardized_request,
            request_hash=None,
        )
    )

    decoy.when(
        action_dispatcher.dispatch(  # type: ignore[func-returns-value]
            QueueCommandAction(
                command_id="command-id-validated",
                created_at=created_at,
                request=standardized_request,
                request_hash=None,
            )
        )
    ).then_do(_stub_queued)

    decoy.when(
        await state_store.wait_for(
            condition=state_store.commands.get_command_is_final,
            command_id="command-id",
        ),
    ).then_do(_stub_completed)

    result = await subject.add_and_execute_command(original_request)

    assert result == completed


async def test_add_and_execute_command_wait_for_recovery(
    decoy: Decoy,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
    subject: ProtocolEngine,
) -> None:
    """It should add and execute a command from a request."""
    created_at = datetime(year=2021, month=1, day=1)
    original_request = commands.WaitForResumeCreate(
        params=commands.WaitForResumeParams()
    )
    standardized_request = commands.HomeCreate(params=commands.HomeParams())
    queued = commands.Home(
        id="command-id",
        key="command-key",
        status=commands.CommandStatus.QUEUED,
        createdAt=created_at,
        params=commands.HomeParams(),
    )
    completed = commands.Home(
        id="command-id",
        key="command-key",
        status=commands.CommandStatus.SUCCEEDED,
        createdAt=created_at,
        params=commands.HomeParams(),
    )

    robot_type: RobotType = "OT-3 Standard"
    decoy.when(state_store.config).then_return(
        Config(robot_type=robot_type, deck_type=DeckType.OT3_STANDARD)
    )

    decoy.when(
        slot_standardization.standardize_command(original_request, robot_type)
    ).then_return(standardized_request)

    decoy.when(model_utils.generate_id()).then_return("command-id")
    decoy.when(model_utils.get_timestamp()).then_return(created_at)

    def _stub_queued(*_a: object, **_k: object) -> None:
        decoy.when(state_store.commands.get("command-id")).then_return(queued)

    def _stub_completed(*_a: object, **_k: object) -> bool:
        decoy.when(state_store.commands.get("command-id")).then_return(completed)
        return True

    decoy.when(
        state_store.commands.validate_action_allowed(
            QueueCommandAction(
                command_id="command-id",
                created_at=created_at,
                request=standardized_request,
                request_hash=None,
            )
        )
    ).then_return(
        QueueCommandAction(
            command_id="command-id-validated",
            created_at=created_at,
            request=standardized_request,
            request_hash=None,
        )
    )

    decoy.when(
        action_dispatcher.dispatch(  # type: ignore[func-returns-value]
            QueueCommandAction(
                command_id="command-id-validated",
                created_at=created_at,
                request=standardized_request,
                request_hash=None,
            )
        )
    ).then_do(_stub_queued)

    decoy.when(
        await state_store.wait_for(
            condition=state_store.commands.get_command_is_final,
            command_id="command-id",
        ),
    ).then_do(_stub_completed)

    result = await subject.add_and_execute_command_wait_for_recovery(original_request)
    assert result == completed
    decoy.verify(
        await state_store.wait_for_not(
            state_store.commands.get_recovery_in_progress_for_command,
            "command-id",
        )
    )


def test_play(
    decoy: Decoy,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
    hardware_api: HardwareControlAPI,
    subject: ProtocolEngine,
) -> None:
    """It should be able to start executing queued commands."""
    decoy.when(model_utils.get_timestamp()).then_return(
        datetime(year=2021, month=1, day=1)
    )
    decoy.when(
        state_store.commands.validate_action_allowed(
            PlayAction(requested_at=datetime(year=2021, month=1, day=1))
        ),
    ).then_return(PlayAction(requested_at=datetime(year=2022, month=2, day=2)))

    subject.play()

    decoy.verify(
        action_dispatcher.dispatch(
            PlayAction(requested_at=datetime(year=2022, month=2, day=2))
        ),
        hardware_api.resume(HardwarePauseType.PAUSE),
    )


def test_play_blocked_by_door(
    decoy: Decoy,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
    hardware_api: HardwareControlAPI,
    subject: ProtocolEngine,
) -> None:
    """It should not pause instead of resuming the hardware if blocked by door."""
    decoy.when(model_utils.get_timestamp()).then_return(
        datetime(year=2021, month=1, day=1)
    )
    decoy.when(
        state_store.commands.validate_action_allowed(
            PlayAction(requested_at=datetime(year=2021, month=1, day=1))
        ),
    ).then_return(PlayAction(requested_at=datetime(year=2022, month=2, day=2)))
    decoy.when(state_store.commands.get_is_door_blocking()).then_return(True)

    subject.play()

    decoy.verify(hardware_api.resume(HardwarePauseType.PAUSE), times=0)
    decoy.verify(
        action_dispatcher.dispatch(
            PlayAction(requested_at=datetime(year=2022, month=2, day=2))
        ),
        hardware_api.pause(HardwarePauseType.PAUSE),
    )


def test_pause(
    decoy: Decoy,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    hardware_api: HardwareControlAPI,
    subject: ProtocolEngine,
) -> None:
    """It should be able to pause executing queued commands."""
    expected_action = PauseAction(source=PauseSource.CLIENT)

    decoy.when(
        state_store.commands.validate_action_allowed(expected_action),
    ).then_return(expected_action)

    subject.request_pause()

    decoy.verify(
        action_dispatcher.dispatch(expected_action),
        hardware_api.pause(HardwarePauseType.PAUSE),
    )


@pytest.mark.parametrize("reconcile_false_positive", [True, False])
def test_resume_from_recovery(
    decoy: Decoy,
    state_store: StateStore,
    action_dispatcher: ActionDispatcher,
    subject: ProtocolEngine,
    reconcile_false_positive: bool,
) -> None:
    """It should dispatch a ResumeFromRecoveryAction."""
    decoy.when(state_store.commands.get_state_update_for_false_positive()).then_return(
        sentinel.state_update_for_false_positive
    )
    empty_state_update = StateUpdate()

    expected_action = ResumeFromRecoveryAction(
        sentinel.state_update_for_false_positive
        if reconcile_false_positive
        else empty_state_update
    )

    decoy.when(
        state_store.commands.validate_action_allowed(expected_action)
    ).then_return(expected_action)

    subject.resume_from_recovery(reconcile_false_positive)

    decoy.verify(action_dispatcher.dispatch(expected_action))


@pytest.mark.parametrize("drop_tips_after_run", [True, False])
@pytest.mark.parametrize("set_run_status", [True, False])
@pytest.mark.parametrize(
    argnames=["post_run_hardware_state", "expected_halt_disengage"],
    argvalues=[
        (PostRunHardwareState.HOME_AND_STAY_ENGAGED, True),
        (PostRunHardwareState.HOME_THEN_DISENGAGE, True),
        (PostRunHardwareState.STAY_ENGAGED_IN_PLACE, False),
        (PostRunHardwareState.DISENGAGE_IN_PLACE, True),
    ],
)
async def test_finish(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    plugin_starter: PluginStarter,
    queue_worker: QueueWorker,
    subject: ProtocolEngine,
    hardware_stopper: HardwareStopper,
    drop_tips_after_run: bool,
    set_run_status: bool,
    post_run_hardware_state: PostRunHardwareState,
    expected_halt_disengage: bool,
    model_utils: ModelUtils,
    state_store: StateStore,
    door_watcher: DoorWatcher,
) -> None:
    """It should be able to gracefully tell the engine it's done."""
    completed_at = datetime(2021, 1, 1, 0, 0)

    decoy.when(state_store.commands.get_is_stopped_by_async_error()).then_return(False)
    decoy.when(model_utils.get_timestamp()).then_return(completed_at)

    await subject.finish(
        drop_tips_after_run=drop_tips_after_run,
        set_run_status=set_run_status,
        post_run_hardware_state=post_run_hardware_state,
    )

    decoy.verify(
        action_dispatcher.dispatch(FinishAction(set_run_status=set_run_status)),
        await queue_worker.join(),
        await hardware_stopper.do_halt(
            disengage_before_stopping=expected_halt_disengage
        ),
        door_watcher.stop(),
        await hardware_stopper.do_stop_and_recover(
            drop_tips_after_run=drop_tips_after_run,
            post_run_hardware_state=post_run_hardware_state,
        ),
        await plugin_starter.stop(),
        action_dispatcher.dispatch(
            HardwareStoppedAction(completed_at=completed_at, finish_error_details=None)
        ),
    )


async def test_finish_with_defaults(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    subject: ProtocolEngine,
    hardware_stopper: HardwareStopper,
    state_store: StateStore,
) -> None:
    """It should be able to gracefully tell the engine it's done."""
    decoy.when(state_store.commands.get_is_stopped_by_async_error()).then_return(False)
    await subject.finish()

    decoy.verify(
        action_dispatcher.dispatch(FinishAction(set_run_status=True)),
        await hardware_stopper.do_halt(disengage_before_stopping=True),
        await hardware_stopper.do_stop_and_recover(
            drop_tips_after_run=True,
            post_run_hardware_state=PostRunHardwareState.HOME_AND_STAY_ENGAGED,
        ),
    )


@pytest.mark.parametrize(
    argnames=["stopped_by_estop", "expected_drop_tips", "expected_end_state"],
    argvalues=[
        (True, False, PostRunHardwareState.DISENGAGE_IN_PLACE),
        (False, True, PostRunHardwareState.HOME_AND_STAY_ENGAGED),
    ],
)
async def test_finish_with_error(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    plugin_starter: PluginStarter,
    queue_worker: QueueWorker,
    model_utils: ModelUtils,
    subject: ProtocolEngine,
    hardware_stopper: HardwareStopper,
    door_watcher: DoorWatcher,
    state_store: StateStore,
    stopped_by_estop: bool,
    expected_drop_tips: bool,
    expected_end_state: PostRunHardwareState,
) -> None:
    """It should be able to tell the engine it's finished because of an error."""
    error = RuntimeError("oh no")
    expected_error_details = FinishErrorDetails(
        error_id="error-id",
        created_at=datetime(year=2021, month=1, day=1),
        error=error,
    )

    decoy.when(state_store.commands.get_is_stopped_by_async_error()).then_return(
        stopped_by_estop
    )
    decoy.when(model_utils.generate_id()).then_return("error-id")
    decoy.when(model_utils.get_timestamp()).then_return(
        datetime(year=2021, month=1, day=1), datetime(year=2022, month=2, day=2)
    )

    await subject.finish(error=error)

    decoy.verify(
        action_dispatcher.dispatch(
            FinishAction(error_details=expected_error_details, set_run_status=True)
        ),
        await queue_worker.join(),
        await hardware_stopper.do_halt(disengage_before_stopping=True),
        door_watcher.stop(),
        await hardware_stopper.do_stop_and_recover(
            drop_tips_after_run=expected_drop_tips,
            post_run_hardware_state=expected_end_state,
        ),
        await plugin_starter.stop(),
        action_dispatcher.dispatch(
            HardwareStoppedAction(
                completed_at=datetime(year=2022, month=2, day=2),
                finish_error_details=None,
            )
        ),
    )


async def test_finish_with_estop_error_will_not_drop_tip_and_home(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    plugin_starter: PluginStarter,
    queue_worker: QueueWorker,
    model_utils: ModelUtils,
    subject: ProtocolEngine,
    hardware_stopper: HardwareStopper,
    door_watcher: DoorWatcher,
) -> None:
    """It should be able to tell the engine it's finished because of an error and will not drop tip and home."""
    error = ProtocolCommandFailedError(
        original_error=ErrorOccurrence.model_construct(  # type: ignore[call-arg]
            wrappedErrors=[
                ErrorOccurrence.model_construct(errorCode="3008")  # type: ignore[call-arg]
            ]
        )
    )
    expected_error_details = FinishErrorDetails(
        error_id="error-id",
        created_at=datetime(year=2021, month=1, day=1),
        error=error,
    )

    decoy.when(model_utils.generate_id()).then_return("error-id")
    decoy.when(model_utils.get_timestamp()).then_return(
        datetime(year=2021, month=1, day=1), datetime(year=2022, month=2, day=2)
    )

    await subject.finish(error=error)

    decoy.verify(
        action_dispatcher.dispatch(
            FinishAction(error_details=expected_error_details, set_run_status=True)
        ),
        await queue_worker.join(),
        await hardware_stopper.do_halt(disengage_before_stopping=True),
        door_watcher.stop(),
        await hardware_stopper.do_stop_and_recover(
            drop_tips_after_run=False,
            post_run_hardware_state=PostRunHardwareState.DISENGAGE_IN_PLACE,
        ),
        await plugin_starter.stop(),
        action_dispatcher.dispatch(
            HardwareStoppedAction(
                completed_at=datetime(year=2022, month=2, day=2),
                finish_error_details=None,
            )
        ),
    )


async def test_finish_stops_hardware_if_queue_worker_join_fails(
    decoy: Decoy,
    queue_worker: QueueWorker,
    hardware_stopper: HardwareStopper,
    door_watcher: DoorWatcher,
    action_dispatcher: ActionDispatcher,
    plugin_starter: PluginStarter,
    subject: ProtocolEngine,
    model_utils: ModelUtils,
    state_store: StateStore,
) -> None:
    """It should be able to stop the engine."""
    exception = RuntimeError("oh no")
    decoy.when(
        await queue_worker.join(),  # type: ignore[func-returns-value]
    ).then_raise(exception)

    decoy.when(state_store.commands.get_is_stopped_by_async_error()).then_return(False)

    error_id = "error-id"
    completed_at = datetime(2021, 1, 1, 0, 0)

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(completed_at)

    await subject.finish()

    decoy.verify(
        action_dispatcher.dispatch(FinishAction()),
        # await _get_queue_worker.join() should be called, and should raise, here.
        # We can't verify that step in the sequence here because of a Decoy limitation.
        await hardware_stopper.do_halt(disengage_before_stopping=True),
        door_watcher.stop(),
        await hardware_stopper.do_stop_and_recover(
            drop_tips_after_run=True,
            post_run_hardware_state=PostRunHardwareState.HOME_AND_STAY_ENGAGED,
        ),
        await plugin_starter.stop(),
        action_dispatcher.dispatch(
            HardwareStoppedAction(
                completed_at=completed_at,
                finish_error_details=FinishErrorDetails(
                    error=exception, error_id="error-id", created_at=completed_at
                ),
            )
        ),
    )


async def test_wait_until_complete(
    decoy: Decoy,
    state_store: StateStore,
    subject: ProtocolEngine,
) -> None:
    """It should be able to stop the engine after waiting for commands to complete."""
    await subject.wait_until_complete()

    decoy.verify(
        await state_store.wait_for(
            condition=state_store.commands.get_all_commands_final
        ),
        state_store.commands.raise_fatal_command_error(),
    )


async def test_stop(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    queue_worker: QueueWorker,
    hardware_stopper: HardwareStopper,
    state_store: StateStore,
    subject: ProtocolEngine,
) -> None:
    """It should be able to stop the engine and run execution."""
    expected_action = StopAction()

    decoy.when(
        state_store.commands.validate_action_allowed(expected_action),
    ).then_return(expected_action)

    await subject.request_stop()

    decoy.verify(
        action_dispatcher.dispatch(expected_action),
        queue_worker.cancel(),
    )


async def test_stop_for_legacy_core_protocols(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    queue_worker: QueueWorker,
    hardware_stopper: HardwareStopper,
    hardware_api: HardwareControlAPI,
    state_store: StateStore,
    subject: ProtocolEngine,
) -> None:
    """It should be able to stop the engine & run execution and cancel movement tasks."""
    expected_action = StopAction()

    decoy.when(
        state_store.commands.validate_action_allowed(expected_action),
    ).then_return(expected_action)

    decoy.when(hardware_api.is_movement_execution_taskified()).then_return(True)

    await subject.request_stop()

    decoy.verify(
        action_dispatcher.dispatch(expected_action),
        queue_worker.cancel(),
        await hardware_api.cancel_execution_and_running_tasks(),
    )


async def test_async_module_error_stops_on_match(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    queue_worker: QueueWorker,
    state_store: StateStore,
    subject: ProtocolEngine,
) -> None:
    """It should be stop the engine if a matching module exists."""
    module_model = ModuleModel.THERMOCYCLER_MODULE_V1
    serial = "hello"
    expected_action = StopAction(from_asynchronous_error=True)
    validated_action = sentinel.validated_action
    decoy.when(
        state_store.commands.validate_action_allowed(expected_action),
    ).then_return(validated_action)
    decoy.when(
        state_store.modules.get_has_module_probably_matching_hardware_details(
            module_model, serial
        )
    ).then_return(True)

    assert await subject.async_module_error(module_model, serial) is True

    decoy.verify(
        action_dispatcher.dispatch(action=validated_action),
        queue_worker.cancel(),
    )


async def test_async_module_error_noops_on_no_match(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    queue_worker: QueueWorker,
    state_store: StateStore,
    subject: ProtocolEngine,
) -> None:
    """It should be stop the engine if a matching module exists."""
    module_model = ModuleModel.THERMOCYCLER_MODULE_V1
    serial = "hello"
    validated_action = sentinel.validated_action
    decoy.when(
        state_store.modules.get_has_module_probably_matching_hardware_details(
            module_model, serial
        )
    ).then_return(False)

    assert await subject.async_module_error(module_model, serial) is False

    decoy.verify(
        action_dispatcher.dispatch(action=validated_action),
        queue_worker.cancel(),
        times=0,
    )


async def test_async_module_error_noops_if_invalid(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    queue_worker: QueueWorker,
    state_store: StateStore,
    subject: ProtocolEngine,
) -> None:
    """It should no-op if a stop is invalid right now.."""
    module_model = ModuleModel.THERMOCYCLER_MODULE_V1
    serial = "hello"
    expected_action = StopAction(from_asynchronous_error=True)
    decoy.when(
        state_store.modules.get_has_module_probably_matching_hardware_details(
            module_model, serial
        )
    ).then_return(True)
    decoy.when(
        state_store.commands.validate_action_allowed(expected_action),
    ).then_raise(RuntimeError("unable to stop; this machine craves flesh"))

    assert (
        await subject.async_module_error(module_model, serial) is True
    )  # Should not raise, should act as-if it worked

    decoy.verify(
        action_dispatcher.dispatch(expected_action),
        times=0,
    )
    decoy.verify(
        queue_worker.cancel(),
        ignore_extra_args=True,
        times=0,
    )


async def test_estop(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    queue_worker: QueueWorker,
    state_store: StateStore,
    subject: ProtocolEngine,
) -> None:
    """It should be able to stop the engine."""
    expected_action = StopAction(from_asynchronous_error=True)
    validated_action = sentinel.validated_action
    decoy.when(
        state_store.commands.validate_action_allowed(expected_action),
    ).then_return(validated_action)

    subject.estop()

    decoy.verify(
        action_dispatcher.dispatch(action=validated_action),
        queue_worker.cancel(),
    )


async def test_estop_noops_if_invalid(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    queue_worker: QueueWorker,
    state_store: StateStore,
    subject: ProtocolEngine,
) -> None:
    """It should no-op if a stop is invalid right now.."""
    expected_action = StopAction(from_asynchronous_error=True)
    decoy.when(
        state_store.commands.validate_action_allowed(expected_action),
    ).then_raise(RuntimeError("unable to stop; this machine craves flesh"))

    subject.estop()  # Should not raise.

    decoy.verify(
        action_dispatcher.dispatch(expected_action),
        times=0,
    )
    decoy.verify(
        queue_worker.cancel(),
        ignore_extra_args=True,
        times=0,
    )


def test_add_plugin(
    decoy: Decoy,
    plugin_starter: PluginStarter,
    subject: ProtocolEngine,
) -> None:
    """It should add a plugin to the PluginStarter."""
    plugin = decoy.mock(cls=AbstractPlugin)

    subject.add_plugin(plugin)

    decoy.verify(plugin_starter.start(plugin))


def test_add_legacy_labware_offset(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
    state_store: StateStore,
    subject: ProtocolEngine,
) -> None:
    """It should have the labware offset request resolved and added to state."""
    request = LegacyLabwareOffsetCreate(
        definitionUri="definition-uri",
        location=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )

    standardized_request = LabwareOffsetCreateInternal(
        definitionUri="standardized-definition-uri",
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="2")
        ],
        legacyLocation=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_2),
        vector=LabwareOffsetVector(x=2, y=3, z=4),
    )

    id = "labware-offset-id"

    created_at = datetime(year=2021, month=11, day=15)

    expected_result = LabwareOffset(
        id=id,
        createdAt=created_at,
        definitionUri=standardized_request.definitionUri,
        location=standardized_request.legacyLocation,
        locationSequence=standardized_request.locationSequence,
        vector=standardized_request.vector,
    )

    robot_type: RobotType = "OT-3 Standard"
    decoy.when(state_store.config).then_return(
        Config(robot_type=robot_type, deck_type=DeckType.OT3_STANDARD)
    )
    decoy.when(state_store.addressable_areas.deck_definition).then_return(
        cast(DeckDefinitionV5, {})
    )
    decoy.when(
        labware_offset_standardization.standardize_labware_offset_create(
            request, robot_type, cast(DeckDefinitionV5, {})
        )
    ).then_return(standardized_request)
    decoy.when(model_utils.generate_id()).then_return(id)
    decoy.when(model_utils.get_timestamp()).then_return(created_at)
    decoy.when(
        state_store.labware.get_labware_offset(labware_offset_id=id)
    ).then_return(expected_result)

    result = subject.add_labware_offset(
        request=LegacyLabwareOffsetCreate(
            definitionUri="definition-uri",
            location=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_1),
            vector=LabwareOffsetVector(x=1, y=2, z=3),
        )
    )

    assert result == expected_result

    decoy.verify(
        action_dispatcher.dispatch(
            AddLabwareOffsetAction(
                labware_offset_id=id,
                created_at=created_at,
                request=standardized_request,
            )
        )
    )


def test_add_labware_offset(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
    state_store: StateStore,
    subject: ProtocolEngine,
) -> None:
    """It should have the labware offset request resolved and added to state."""
    request = LabwareOffsetCreate(
        definitionUri="definition-uri",
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="1")
        ],
        vector=LabwareOffsetVector(x=1, y=2, z=3),
    )

    standardized_request = LabwareOffsetCreateInternal(
        definitionUri="standardized-definition-uri",
        locationSequence=[
            OnAddressableAreaOffsetLocationSequenceComponent(addressableAreaName="3")
        ],
        legacyLocation=LegacyLabwareOffsetLocation(slotName=DeckSlotName.SLOT_3),
        vector=LabwareOffsetVector(x=2, y=5, z=6),
    )

    id = "labware-offset-id"

    created_at = datetime(year=2021, month=11, day=15)

    expected_result = LabwareOffset(
        id=id,
        createdAt=created_at,
        definitionUri=standardized_request.definitionUri,
        location=standardized_request.legacyLocation,
        locationSequence=standardized_request.locationSequence,
        vector=standardized_request.vector,
    )

    robot_type: RobotType = "OT-3 Standard"
    decoy.when(state_store.config).then_return(
        Config(robot_type=robot_type, deck_type=DeckType.OT3_STANDARD)
    )
    decoy.when(state_store.addressable_areas.deck_definition).then_return(
        cast(DeckDefinitionV5, {})
    )
    decoy.when(
        labware_offset_standardization.standardize_labware_offset_create(
            request, robot_type, cast(DeckDefinitionV5, {})
        )
    ).then_return(standardized_request)
    decoy.when(model_utils.generate_id()).then_return(id)
    decoy.when(model_utils.get_timestamp()).then_return(created_at)
    decoy.when(
        state_store.labware.get_labware_offset(labware_offset_id=id)
    ).then_return(expected_result)

    result = subject.add_labware_offset(
        request=LabwareOffsetCreate(
            definitionUri="definition-uri",
            locationSequence=[
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName="1"
                )
            ],
            vector=LabwareOffsetVector(x=1, y=2, z=3),
        )
    )

    assert result == expected_result

    decoy.verify(
        action_dispatcher.dispatch(
            AddLabwareOffsetAction(
                labware_offset_id=id,
                created_at=created_at,
                request=standardized_request,
            )
        )
    )


def test_add_labware_definition(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    state_store: StateStore,
    subject: ProtocolEngine,
    well_plate_def: LabwareDefinition,
) -> None:
    """It should dispatch an AddLabwareDefinition action."""

    def _stub_get_definition_uri(*args: Any, **kwargs: Any) -> None:
        decoy.when(
            state_store.labware.get_uri_from_definition(well_plate_def)
        ).then_return(LabwareUri("some/definition/uri"))

    decoy.when(
        action_dispatcher.dispatch(  # type: ignore[func-returns-value]
            AddLabwareDefinitionAction(definition=well_plate_def)
        )
    ).then_do(_stub_get_definition_uri)

    result = subject.add_labware_definition(well_plate_def)

    assert result == "some/definition/uri"


def test_add_camera_settings(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    subject: ProtocolEngine,
) -> None:
    """It should dispatch an AddCameraSettingsAction action."""
    settings = CameraSettings(
        cameraEnabled=True, liveStreamEnabled=True, errorRecoveryCameraEnabled=True
    )
    decoy.when(subject.state_view.camera.get_enablement_settings()).then_return(
        settings
    )
    subject.add_camera_enablement_settings(settings)
    decoy.verify(
        action_dispatcher.dispatch(
            AddCameraSettingsAction(enablement_settings=settings)
        )
    )


def test_add_addressable_area(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    subject: ProtocolEngine,
) -> None:
    """It should dispatch an AddAddressableArea action."""
    subject.add_addressable_area(addressable_area_name="my_funky_area")

    decoy.verify(
        action_dispatcher.dispatch(
            AddAddressableAreaAction(addressable_area_name="my_funky_area")
        )
    )


def test_add_liquid(
    decoy: Decoy,
    action_dispatcher: ActionDispatcher,
    subject: ProtocolEngine,
    state_store: StateStore,
) -> None:
    """It should dispatch an AddLiquidAction action."""
    liquid_obj = Liquid(id="water-id", displayName="water", description="water desc")
    decoy.when(
        state_store.liquid.validate_liquid_allowed(liquid=liquid_obj)
    ).then_return(liquid_obj)
    subject.add_liquid(
        id="water-id", name="water", description="water desc", color=None
    )

    decoy.verify(action_dispatcher.dispatch(AddLiquidAction(liquid=liquid_obj)))


async def test_use_attached_temp_and_mag_modules(
    decoy: Decoy,
    module_data_provider: ModuleDataProvider,
    action_dispatcher: ActionDispatcher,
    subject: ProtocolEngine,
    tempdeck_v1_def: ModuleDefinition,
    magdeck_v2_def: ModuleDefinition,
) -> None:
    """It should be able to load attached hardware modules directly into state."""
    mod_1 = decoy.mock(cls=TempDeck)
    mod_2 = decoy.mock(cls=MagDeck)

    decoy.when(mod_1.device_info).then_return({"serial": "serial-1"})
    decoy.when(mod_2.device_info).then_return({"serial": "serial-2"})
    decoy.when(mod_1.model()).then_return("temperatureModuleV1")
    decoy.when(mod_2.model()).then_return("magneticModuleV2")

    decoy.when(mod_1.live_data).then_return({"status": "some-status", "data": {}})
    decoy.when(mod_2.live_data).then_return({"status": "other-status", "data": {}})

    decoy.when(
        module_data_provider.get_definition(ModuleModel.TEMPERATURE_MODULE_V1)
    ).then_return(tempdeck_v1_def)

    decoy.when(
        module_data_provider.get_definition(ModuleModel.MAGNETIC_MODULE_V2)
    ).then_return(magdeck_v2_def)

    await subject.use_attached_modules(
        {
            "module-1": mod_1,
            "module-2": mod_2,
        }
    )

    decoy.verify(
        action_dispatcher.dispatch(
            AddModuleAction(
                module_id="module-1",
                serial_number="serial-1",
                definition=tempdeck_v1_def,
                module_live_data={"status": "some-status", "data": {}},
            )
        ),
        action_dispatcher.dispatch(
            AddModuleAction(
                module_id="module-2",
                serial_number="serial-2",
                definition=magdeck_v2_def,
                module_live_data={"status": "other-status", "data": {}},
            ),
        ),
    )


async def test_set_error_recovery_policy(
    decoy: Decoy, action_dispatcher: ActionDispatcher, subject: ProtocolEngine
) -> None:
    """It should set the error recovery policy by dispatching an action."""
    subject.set_error_recovery_policy(sentinel.new_policy)
    decoy.verify(
        action_dispatcher.dispatch(SetErrorRecoveryPolicyAction(sentinel.new_policy))
    )



"""Tests for Protocol API temperature module contexts."""

import pytest
from decoy import Decoy, matchers

from opentrons.hardware_control.modules import TemperatureStatus
from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_api import MAX_SUPPORTED_VERSION, TemperatureModuleContext
from opentrons.protocol_api.core.common import ProtocolCore, TemperatureModuleCore
from opentrons.protocol_api.core.core_map import LoadedCoreMap
from opentrons.protocol_api.tasks import Task
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import APIVersionError


@pytest.fixture
def mock_core(decoy: Decoy) -> TemperatureModuleCore:
    """Get a mock module implementation core."""
    return decoy.mock(cls=TemperatureModuleCore)


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol implementation core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def mock_core_map(decoy: Decoy) -> LoadedCoreMap:
    """Get a mock LoadedCoreMap."""
    return decoy.mock(cls=LoadedCoreMap)


@pytest.fixture
def mock_broker(decoy: Decoy) -> LegacyBroker:
    """Get a mock command message broker."""
    return decoy.mock(cls=LegacyBroker)


@pytest.fixture
def api_version() -> APIVersion:
    """Get an API version to apply to the interface."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def subject(
    api_version: APIVersion,
    mock_core: TemperatureModuleCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_broker: LegacyBroker,
) -> TemperatureModuleContext:
    """Get a temperature module context with its dependencies mocked out."""
    return TemperatureModuleContext(
        core=mock_core,
        protocol_core=mock_protocol_core,
        core_map=mock_core_map,
        broker=mock_broker,
        api_version=api_version,
    )


def test_set_temperature(
    decoy: Decoy,
    mock_core: TemperatureModuleCore,
    mock_broker: LegacyBroker,
    subject: TemperatureModuleContext,
) -> None:
    """It should set and wait for the temperature via the core."""
    subject.set_temperature(42.0)

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.TEMPDECK_SET_TEMP",
                    "payload": matchers.DictMatching({"celsius": 42.0}),
                }
            ),
        ),
        mock_core.set_target_temperature(42.0),
        mock_core.wait_for_target_temperature(),
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after"}),
        ),
    )


@pytest.mark.parametrize("api_version", [APIVersion(2, 3), APIVersion(2, 27)])
def test_start_set_temperature(
    decoy: Decoy,
    mock_core: TemperatureModuleCore,
    mock_broker: LegacyBroker,
    subject: TemperatureModuleContext,
    api_version: APIVersion,
) -> None:
    """It should set the target temperature via the core and return a task."""
    subject._api_version = api_version
    mock_task = decoy.mock(cls=Task)
    decoy.when(mock_core.set_target_temperature(42.0)).then_return(mock_task._core)
    result = subject.start_set_temperature(42.0)

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.TEMPDECK_SET_TEMP",
                    "payload": matchers.DictMatching({"celsius": 42.0}),
                }
            ),
        ),
        mock_broker.publish("command", matchers.DictMatching({"$": "after"})),
    )

    decoy.verify(
        mock_core.wait_for_target_temperature(), ignore_extra_args=True, times=0
    )
    if api_version >= APIVersion(2, 27):
        assert isinstance(result, Task)
        assert result._core is mock_task._core
        assert result._api_version == api_version
    else:
        assert result is None


@pytest.mark.parametrize("api_version", [APIVersion(2, 2)])
def test_start_set_temperature_api_version_low(
    decoy: Decoy, mock_broker: LegacyBroker, subject: TemperatureModuleContext
) -> None:
    """It should reject if API version is lower than 2.3."""
    with pytest.raises(APIVersionError) as exc_info:
        subject.start_set_temperature(42.0)

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after", "error": exc_info.value}),
        ),
    )


@pytest.mark.parametrize("api_version", [APIVersion(2, 3)])
def test_await_temperature(
    decoy: Decoy,
    mock_core: TemperatureModuleCore,
    mock_broker: LegacyBroker,
    subject: TemperatureModuleContext,
) -> None:
    """It should wait for a specified target temperature."""
    subject.await_temperature(42.0)

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {
                    "$": "before",
                    "name": "command.TEMPDECK_AWAIT_TEMP",
                    "payload": matchers.DictMatching({"celsius": 42.0}),
                }
            ),
        ),
        mock_core.wait_for_target_temperature(42.0),
        mock_broker.publish("command", matchers.DictMatching({"$": "after"})),
    )


@pytest.mark.parametrize("api_version", [APIVersion(2, 2)])
def test_await_temperature_api_version_low(
    decoy: Decoy, mock_broker: LegacyBroker, subject: TemperatureModuleContext
) -> None:
    """It should reject if API version is lower than 2.3."""
    with pytest.raises(APIVersionError) as exc_info:
        subject.await_temperature(42.0)

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching({"$": "after", "error": exc_info.value}),
        ),
    )


def test_deactivate(
    decoy: Decoy,
    mock_core: TemperatureModuleCore,
    mock_broker: LegacyBroker,
    subject: TemperatureModuleContext,
) -> None:
    """It should deactivate the heater."""
    subject.deactivate()

    decoy.verify(
        mock_broker.publish(
            "command",
            matchers.DictMatching(
                {"$": "before", "name": "command.TEMPDECK_DEACTIVATE"}
            ),
        ),
        mock_core.deactivate(),
        mock_broker.publish("command", matchers.DictMatching({"$": "after"})),
    )


def test_get_current_temperature(
    decoy: Decoy, mock_core: TemperatureModuleCore, subject: TemperatureModuleContext
) -> None:
    """It should get the current temperature from the core."""
    decoy.when(mock_core.get_current_temperature()).then_return(42.0)

    result = subject.temperature

    assert result == 42.0


def test_get_target_temperature(
    decoy: Decoy, mock_core: TemperatureModuleCore, subject: TemperatureModuleContext
) -> None:
    """It should get the target temperature from the core."""
    decoy.when(mock_core.get_target_temperature()).then_return(42.0)

    result = subject.target

    assert result == 42.0


@pytest.mark.parametrize("api_version", [APIVersion(2, 3)])
def test_get_status(
    decoy: Decoy, mock_core: TemperatureModuleCore, subject: TemperatureModuleContext
) -> None:
    """It should get the target temperature from the core."""
    decoy.when(mock_core.get_status()).then_return(TemperatureStatus.HEATING)

    result = subject.status

    assert result == "heating"


@pytest.mark.parametrize("api_version", [APIVersion(2, 2)])
def test_get_status_version_low(subject: TemperatureModuleContext) -> None:
    """It should reject if API version is lower than 2.3."""
    with pytest.raises(APIVersionError):
        subject.status


def test_serial_number(
    decoy: Decoy,
    mock_core: TemperatureModuleCore,
    subject: TemperatureModuleContext,
) -> None:
    """It should get the module's unique serial number."""
    decoy.when(mock_core.get_serial_number()).then_return("abc-123")
    result = subject.serial_number
    assert result == "abc-123"



"""Test instrument context simulation."""

from typing import Callable, cast

import pytest
from _pytest.fixtures import SubRequest
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons_shared_data.errors.exceptions import (
    UnexpectedTipAttachError,
    UnexpectedTipRemovalError,
)

from opentrons.protocol_api.core.common import InstrumentCore, LabwareCore
from opentrons.types import Location, Point

# TODO (lc 12-8-2022) Not sure if we plan to keep these tests, but if we do
# we should re-write them to be agnostic to the underlying hardware. Otherwise
# I wouldn't really consider these to be proper unit tests.
pytestmark = pytest.mark.ot2_only


@pytest.fixture(
    params=[
        lazy_fixture("instrument_context"),
        lazy_fixture("simulating_instrument_context"),
    ]
)
def subject(request: SubRequest) -> InstrumentCore:
    return cast(InstrumentCore, request.param)


def test_same_pipette(
    instrument_context: InstrumentCore,
    simulating_instrument_context: InstrumentCore,
) -> None:
    """It should have the same pipette as hardware backed instrument context."""
    assert (
        instrument_context.get_hardware_state()
        == simulating_instrument_context.get_hardware_state()
    )


def test_prepare_to_aspirate_no_tip(subject: InstrumentCore) -> None:
    """It should raise an error if a tip is not attached."""
    with pytest.raises(
        UnexpectedTipRemovalError, match="Cannot perform PREPARE_ASPIRATE"
    ):
        subject.prepare_to_aspirate()


def test_dispense_no_tip(subject: InstrumentCore) -> None:
    """It should raise an error if a tip is not attached."""
    subject.home()
    location = Location(point=Point(1, 2, 3), labware=None)
    with pytest.raises(UnexpectedTipRemovalError, match="Cannot perform DISPENSE"):
        subject.dispense(
            volume=1,
            rate=1,
            flow_rate=1,
            location=location,
            well_core=None,
            in_place=False,
            correction_volume=0,
            push_out=None,
        )


def test_pick_up_tip_no_tip(subject: InstrumentCore, tip_rack: LabwareCore) -> None:
    """It should raise an error if a tip is already attached."""
    tip_core = tip_rack.get_well_core("A1")

    subject.home()
    subject.pick_up_tip(
        location=Location(point=tip_core.get_top(z_offset=0), labware=None),
        well_core=tip_core,
        presses=None,
        increment=None,
        prep_after=False,
    )
    with pytest.raises(UnexpectedTipAttachError):
        subject.pick_up_tip(
            location=Location(point=tip_core.get_top(z_offset=0), labware=None),
            well_core=tip_core,
            presses=None,
            increment=None,
            prep_after=False,
        )


def test_pick_up_tip_prep_after(
    subject: InstrumentCore, labware: LabwareCore, tip_rack: LabwareCore
) -> None:
    """It should not raise an error, regardless of prep_after value."""
    tip_core = tip_rack.get_well_core("A1")

    subject.home()
    subject.pick_up_tip(
        location=Location(point=tip_core.get_top(z_offset=0), labware=None),
        well_core=tip_core,
        presses=None,
        increment=None,
        prep_after=True,
    )
    subject.aspirate(
        location=Location(point=Point(1, 2, 3), labware=None),
        well_core=labware.get_well_core("A1"),
        volume=1,
        rate=1,
        flow_rate=1,
        in_place=False,
        correction_volume=0,
    )
    subject.dispense(
        volume=1,
        rate=1,
        flow_rate=1,
        location=Location(point=Point(2, 2, 3), labware=None),
        well_core=labware.get_well_core("A2"),
        in_place=False,
        correction_volume=0,
        push_out=None,
    )

    subject.drop_tip(location=None, well_core=tip_core, home_after=True)

    # and again, without preparing for aspirate
    subject.pick_up_tip(
        location=Location(point=tip_core.get_top(z_offset=0), labware=None),
        well_core=tip_core,
        presses=None,
        increment=None,
        prep_after=False,
    )
    subject.aspirate(
        location=Location(point=Point(1, 2, 3), labware=None),
        well_core=labware.get_well_core("A1"),
        volume=1,
        rate=1,
        flow_rate=1,
        in_place=False,
        correction_volume=0,
    )
    subject.dispense(
        volume=1,
        rate=1,
        flow_rate=1,
        location=Location(point=Point(2, 2, 3), labware=None),
        well_core=labware.get_well_core("A2"),
        in_place=False,
        correction_volume=0,
        push_out=None,
    )

    subject.drop_tip(location=None, well_core=tip_core, home_after=True)


def test_aspirate_too_much(
    subject: InstrumentCore, labware: LabwareCore, tip_rack: LabwareCore
) -> None:
    """It should raise an error if try to aspirate more than possible."""
    subject.home()
    subject.pick_up_tip(
        location=Location(
            point=tip_rack.get_well_core("A1").get_top(z_offset=0), labware=None
        ),
        well_core=tip_rack.get_well_core("A1"),
        presses=None,
        increment=None,
        prep_after=False,
    )
    subject.prepare_to_aspirate()
    with pytest.raises(
        AssertionError, match="Cannot aspirate more than pipette max volume"
    ):
        subject.aspirate(
            location=Location(point=Point(1, 2, 3), labware=None),
            well_core=labware.get_well_core("A1"),
            volume=subject.get_max_volume() + 1,
            rate=1,
            flow_rate=1,
            in_place=False,
            correction_volume=0,
        )


def test_working_volume(subject: InstrumentCore, tip_rack: LabwareCore) -> None:
    """It should have the correct working volume."""
    subject.home()
    assert subject.get_hardware_state()["working_volume"] == 300
    subject.pick_up_tip(
        location=Location(
            point=tip_rack.get_well_core("A1").get_top(z_offset=0), labware=None
        ),
        well_core=tip_rack.get_well_core("A1"),
        presses=None,
        increment=None,
        prep_after=False,
    )
    assert subject.get_hardware_state()["working_volume"] == 200


@pytest.mark.parametrize(
    argnames=["side_effector"],
    argvalues=[
        [lambda i: None],
        [lambda i: i.set_flow_rate(aspirate=212, dispense=44, blow_out=22)],
        [lambda i: i.set_pipette_speed(aspirate=212, dispense=44, blow_out=22)],
    ],
)
def test_pipette_dict(
    side_effector: Callable[[InstrumentCore], None],
    instrument_context: InstrumentCore,
    simulating_instrument_context: InstrumentCore,
) -> None:
    """It should be the same."""
    side_effector(instrument_context)
    side_effector(simulating_instrument_context)
    assert (
        instrument_context.get_hardware_state()
        == simulating_instrument_context.get_hardware_state()
    )


def _aspirate(i: InstrumentCore, labware: LabwareCore) -> None:
    """pipette dict with tip fixture."""
    i.prepare_to_aspirate()
    i.aspirate(
        location=Location(point=Point(1, 2, 3), labware=None),
        well_core=labware.get_well_core("A1"),
        volume=12,
        rate=10,
        flow_rate=10,
        in_place=False,
        correction_volume=0,
    )


def _aspirate_dispense(i: InstrumentCore, labware: LabwareCore) -> None:
    """pipette dict with tip fixture."""
    i.prepare_to_aspirate()
    i.aspirate(
        location=Location(point=Point(1, 2, 3), labware=None),
        well_core=labware.get_well_core("A1"),
        volume=12,
        rate=10,
        flow_rate=10,
        in_place=False,
        correction_volume=0,
    )
    i.dispense(
        volume=2,
        rate=2,
        flow_rate=2,
        location=Location(point=Point(2, 2, 3), labware=None),
        well_core=labware.get_well_core("A2"),
        in_place=False,
        correction_volume=0,
        push_out=None,
    )


def _aspirate_blowout(i: InstrumentCore, labware: LabwareCore) -> None:
    """pipette dict with tip fixture."""
    i.prepare_to_aspirate()
    i.aspirate(
        location=Location(point=Point(1, 2, 3), labware=None),
        well_core=labware.get_well_core("A1"),
        volume=11,
        rate=13,
        flow_rate=13,
        in_place=False,
        correction_volume=0,
    )
    i.blow_out(
        location=Location(point=Point(1, 2, 3), labware=None),
        well_core=labware.get_well_core("A1"),
        in_place=True,
        flow_rate=789,
    )


@pytest.mark.parametrize(
    argnames=["side_effector"],
    argvalues=[
        [lambda i, l: None],  # noqa: E741
        [_aspirate],
        [_aspirate_dispense],
        [_aspirate_blowout],
    ],
)
def test_pipette_dict_with_tip(
    side_effector: Callable[[InstrumentCore, LabwareCore], None],
    instrument_context: InstrumentCore,
    simulating_instrument_context: InstrumentCore,
    labware: LabwareCore,
    tip_rack: LabwareCore,
) -> None:
    """It should be the same."""
    tip_core = tip_rack.get_well_core("A1")

    # Home first
    instrument_context.home()
    simulating_instrument_context.home()
    # Pickup tip
    instrument_context.pick_up_tip(
        location=Location(point=tip_core.get_top(z_offset=0), labware=None),
        well_core=tip_core,
        presses=3,
        increment=4,
        prep_after=False,
    )
    simulating_instrument_context.pick_up_tip(
        location=Location(point=tip_core.get_top(z_offset=0), labware=None),
        well_core=tip_core,
        presses=3,
        increment=4,
        prep_after=False,
    )

    side_effector(instrument_context, labware)
    side_effector(simulating_instrument_context, labware)
    assert (
        instrument_context.get_hardware_state()
        == simulating_instrument_context.get_hardware_state()
    )

    # Drop tip and compare again
    instrument_context.drop_tip(location=None, well_core=tip_core, home_after=False)
    simulating_instrument_context.drop_tip(
        location=None, well_core=tip_core, home_after=False
    )

    assert (
        instrument_context.get_hardware_state()
        == simulating_instrument_context.get_hardware_state()
    )



import pytest

from opentrons.drivers.types import Temperature
from opentrons.hardware_control.modules.lid_temp_status import LidTemperatureStatus
from opentrons.hardware_control.modules.types import TemperatureStatus


@pytest.mark.parametrize(
    argnames=["temp", "expected"],
    argvalues=[
        # No target
        [Temperature(current=50, target=None), TemperatureStatus.IDLE],
        # At target
        [Temperature(current=50, target=50), TemperatureStatus.HOLDING],
        # Close enough to target
        [
            Temperature(
                current=50 + LidTemperatureStatus.TEMP_THRESHOLD / 2, target=50
            ),
            TemperatureStatus.HOLDING,
        ],
        [
            Temperature(
                current=50 - LidTemperatureStatus.TEMP_THRESHOLD / 2, target=50
            ),
            TemperatureStatus.HOLDING,
        ],
        [
            Temperature(current=50 + LidTemperatureStatus.TEMP_THRESHOLD, target=50),
            TemperatureStatus.HOLDING,
        ],
        [
            Temperature(current=50 - LidTemperatureStatus.TEMP_THRESHOLD, target=50),
            TemperatureStatus.HOLDING,
        ],
        # Above target
        [
            Temperature(
                current=50 + (LidTemperatureStatus.TEMP_THRESHOLD + 1), target=50
            ),
            TemperatureStatus.IDLE,
        ],
        # Below target
        [
            Temperature(
                current=50 - (LidTemperatureStatus.TEMP_THRESHOLD + 1), target=50
            ),
            TemperatureStatus.HEATING,
        ],
    ],
)
def test_lid_temp_status(temp: Temperature, expected: TemperatureStatus) -> None:
    """It should set the status correctly based on temperature reading."""
    status = LidTemperatureStatus()
    assert status.update(temp) == expected
    assert status.status == expected


def test_lid_temp_status_error() -> None:
    """It should be error status when uninitialized ."""
    assert LidTemperatureStatus().status == TemperatureStatus.ERROR



"""Tests for the PythonAndLegacyRunner's LegacyContextPlugin."""

import asyncio
from datetime import datetime
from typing import Callable

import pytest
from anyio import to_thread
from decoy import Decoy, matchers

from opentrons_shared_data.labware.types import (
    LabwareDefinition2 as LabwareDefinition2Dict,
)

from opentrons.legacy_broker import LegacyBroker
from opentrons.legacy_commands.types import (
    CommandMessage as LegacyCommand,
)
from opentrons.legacy_commands.types import (
    PauseMessage,
)
from opentrons.protocol_api.core.legacy.load_info import LabwareLoadInfo, LoadInfo
from opentrons.protocol_engine import (
    StateView,
)
from opentrons.protocol_engine import (
    actions as pe_actions,
)
from opentrons.protocol_engine import (
    commands as pe_commands,
)
from opentrons.protocol_runner.legacy_command_mapper import LegacyCommandMapper
from opentrons.protocol_runner.legacy_context_plugin import LegacyContextPlugin
from opentrons.types import DeckSlotName
from opentrons.util.broker import ReadOnlyBroker


@pytest.fixture
def mock_legacy_broker(decoy: Decoy) -> LegacyBroker:
    """Get a mocked out `broker: LegacyBroker` dependency."""
    return decoy.mock(cls=LegacyBroker)


@pytest.fixture
def mock_equipment_broker(decoy: Decoy) -> ReadOnlyBroker[LoadInfo]:
    """Get a mocked out `equipment_broker: Broker` dependency."""
    return decoy.mock(cls=ReadOnlyBroker[LoadInfo])


@pytest.fixture
def mock_legacy_command_mapper(decoy: Decoy) -> LegacyCommandMapper:
    """Get a mocked out LegacyCommandMapper dependency."""
    return decoy.mock(cls=LegacyCommandMapper)


@pytest.fixture
def mock_state_view(decoy: Decoy) -> StateView:
    """Get a mock StateView."""
    return decoy.mock(cls=StateView)


@pytest.fixture
def mock_action_dispatcher(decoy: Decoy) -> pe_actions.ActionDispatcher:
    """Get a mock ActionDispatcher."""
    return decoy.mock(cls=pe_actions.ActionDispatcher)


@pytest.fixture
async def subject(
    mock_legacy_broker: LegacyBroker,
    mock_equipment_broker: ReadOnlyBroker[LoadInfo],
    mock_legacy_command_mapper: LegacyCommandMapper,
    mock_state_view: StateView,
    mock_action_dispatcher: pe_actions.ActionDispatcher,
) -> LegacyContextPlugin:
    """Get a configured LegacyContextPlugin with its dependencies mocked out."""
    plugin = LegacyContextPlugin(
        engine_loop=asyncio.get_running_loop(),
        broker=mock_legacy_broker,
        equipment_broker=mock_equipment_broker,
        legacy_command_mapper=mock_legacy_command_mapper,
    )
    plugin._configure(state=mock_state_view, action_dispatcher=mock_action_dispatcher)
    return plugin


class _ContextManager:
    """A placeholder in the shape of a context manager, to pass as a template to `decoy.mock()`."""

    def __enter__(self) -> None:
        raise NotImplementedError()

    def __exit__(self, type: object, value: object, traceback: object) -> None:
        raise NotImplementedError()


async def test_broker_subscribe_unsubscribe(
    decoy: Decoy,
    mock_legacy_broker: LegacyBroker,
    mock_equipment_broker: ReadOnlyBroker[LoadInfo],
    subject: LegacyContextPlugin,
) -> None:
    """It should subscribe to the brokers on setup and unsubscribe on teardown."""
    command_broker_unsubscribe: Callable[[], None] = decoy.mock(
        name="command_broker_unsubscribe"
    )
    equipment_broker_subscription_context = decoy.mock(cls=_ContextManager)

    decoy.when(
        mock_legacy_broker.subscribe(topic="command", handler=matchers.Anything())
    ).then_return(command_broker_unsubscribe)

    decoy.when(
        mock_equipment_broker.subscribed(callback=matchers.Anything())
    ).then_return(equipment_broker_subscription_context)

    subject.setup()
    await subject.teardown()

    decoy.verify(command_broker_unsubscribe())

    decoy.verify(
        [
            equipment_broker_subscription_context.__enter__(),  # type: ignore[func-returns-value]
            equipment_broker_subscription_context.__exit__(None, None, None),  # type: ignore[func-returns-value]
        ]
    )


async def test_command_broker_messages(
    decoy: Decoy,
    mock_legacy_broker: LegacyBroker,
    mock_equipment_broker: ReadOnlyBroker[LoadInfo],
    mock_legacy_command_mapper: LegacyCommandMapper,
    mock_action_dispatcher: pe_actions.ActionDispatcher,
    subject: LegacyContextPlugin,
) -> None:
    """It should dispatch commands from command broker messages."""
    # Capture the function that the plugin sets up as its command broker callback.
    # Also, ensure that all subscribe calls return an actual unsubscribe callable
    # (instead of Decoy's default `None`) so subject.teardown() works.
    command_handler_captor = matchers.Captor()
    decoy.when(
        mock_legacy_broker.subscribe(topic="command", handler=command_handler_captor)
    ).then_return(decoy.mock(name="command_broker_unsubscribe"))
    decoy.when(
        mock_equipment_broker.subscribed(callback=matchers.Anything())
    ).then_enter_with(None)

    subject.setup()

    handler: Callable[[LegacyCommand], None] = command_handler_captor.value

    legacy_command: PauseMessage = {
        "$": "before",
        "id": "message-id",
        "name": "command.PAUSE",
        "payload": {"userMessage": "hello", "text": "hello"},
        "error": None,
    }
    engine_command = pe_commands.Custom(
        id="command-id",
        key="command-key",
        status=pe_commands.CommandStatus.RUNNING,
        createdAt=datetime(year=2021, month=1, day=1),
        params=pe_commands.CustomParams(message="hello"),  # type: ignore[call-arg]
    )

    decoy.when(
        mock_legacy_command_mapper.map_command(command=legacy_command)
    ).then_return([pe_actions.SucceedCommandAction(engine_command)])

    await to_thread.run_sync(handler, legacy_command)

    await subject.teardown()

    decoy.verify(
        mock_action_dispatcher.dispatch(pe_actions.SucceedCommandAction(engine_command))
    )


async def test_equipment_broker_messages(
    decoy: Decoy,
    mock_legacy_broker: LegacyBroker,
    mock_equipment_broker: ReadOnlyBroker[LoadInfo],
    mock_legacy_command_mapper: LegacyCommandMapper,
    mock_action_dispatcher: pe_actions.ActionDispatcher,
    subject: LegacyContextPlugin,
    minimal_labware_def: LabwareDefinition2Dict,
) -> None:
    """It should dispatch commands from equipment broker messages."""
    # Capture the function that the plugin sets up as its labware load callback.
    labware_handler_captor = matchers.Captor()
    decoy.when(
        mock_legacy_broker.subscribe(topic="command", handler=matchers.Anything())
    ).then_return(decoy.mock(name="command_broker_unsubscribe"))
    decoy.when(
        mock_equipment_broker.subscribed(callback=labware_handler_captor)
    ).then_enter_with(None)

    subject.setup()

    handler: Callable[[LabwareLoadInfo], None] = labware_handler_captor.value

    load_info = LabwareLoadInfo(
        labware_definition=minimal_labware_def,
        labware_namespace="some_namespace",
        labware_load_name="some_load_name",
        labware_display_name=None,
        labware_version=123,
        deck_slot=DeckSlotName.SLOT_1,
        on_module=False,
        offset_id=None,
    )

    engine_command = pe_commands.Custom(
        id="command-id",
        key="command-key",
        status=pe_commands.CommandStatus.RUNNING,
        createdAt=datetime(year=2021, month=1, day=1),
        params=pe_commands.CustomParams(message="hello"),  # type: ignore[call-arg]
    )

    decoy.when(
        mock_legacy_command_mapper.map_equipment_load(load_info=load_info)
    ).then_return([pe_actions.SucceedCommandAction(command=engine_command)])

    await to_thread.run_sync(handler, load_info)

    await subject.teardown()

    decoy.verify(
        mock_action_dispatcher.dispatch(
            pe_actions.SucceedCommandAction(command=engine_command)
        ),
    )



"""Support for create_protocol_engine module."""

from ..hardware_control import HardwareControlAPI
from . import ProtocolEngine
from opentrons.protocol_engine.resources.camera_provider import CameraProvider
from opentrons.protocol_runner import RunOrchestrator, protocol_runner


def create_run_orchestrator(
    hardware_api: HardwareControlAPI,
    protocol_engine: ProtocolEngine,
    camera_provider: CameraProvider,
) -> RunOrchestrator:
    """Create a RunOrchestrator instance."""
    return RunOrchestrator(
        protocol_engine=protocol_engine,
        hardware_api=hardware_api,
        camera_provider=camera_provider,
        setup_runner=protocol_runner.LiveRunner(
            protocol_engine=protocol_engine,
            hardware_api=hardware_api,
        ),
        fixit_runner=protocol_runner.LiveRunner(
            protocol_engine=protocol_engine,
            hardware_api=hardware_api,
        ),
        protocol_live_runner=protocol_runner.LiveRunner(
            protocol_engine=protocol_engine,
            hardware_api=hardware_api,
        ),
    )



import io
import zipfile
from typing import Any, Callable

import pytest

from opentrons.protocols import bundle


def test_parse_bundle_no_root_files(get_bundle_fixture: Callable[..., Any]) -> None:
    fixture = get_bundle_fixture("no_root_files_bundle")
    with pytest.raises(
        RuntimeError, match="No files found in ZIP file's root directory"
    ):
        buf = io.BytesIO(fixture["binary_zipfile"])
        buf.seek(0)
        zf = zipfile.ZipFile(buf)
        bundle.extract_bundle(zf)


def test_parse_bundle_no_entrypoint_protocol(
    get_bundle_fixture: Callable[..., Any],
) -> None:
    fixture = get_bundle_fixture("no_entrypoint_protocol_bundle")
    with pytest.raises(RuntimeError, match="Bundled protocol should have a"):
        buf = io.BytesIO(fixture["binary_zipfile"])
        buf.seek(0)
        zf = zipfile.ZipFile(buf)
        bundle.extract_bundle(zf)


def test_parse_bundle_conflicting_labware(
    get_bundle_fixture: Callable[..., Any],
) -> None:
    fixture = get_bundle_fixture("conflicting_labware_bundle")
    with pytest.raises(RuntimeError, match="Conflicting labware in bundle"):
        buf = io.BytesIO(fixture["binary_zipfile"])
        buf.seek(0)
        zf = zipfile.ZipFile(buf)
        bundle.extract_bundle(zf)


def test_write_bundle(get_bundle_fixture: Callable[..., Any]) -> None:
    fixture = get_bundle_fixture("simple_bundle")
    buf = io.BytesIO(fixture["binary_zipfile"])
    buf.seek(0)
    zf = zipfile.ZipFile(buf)
    original_contents = bundle.extract_bundle(zf)
    t = io.BytesIO()
    bundle.create_bundle(original_contents, t)
    zf2 = zipfile.ZipFile(t)
    new_contents = bundle.extract_bundle(zf2)
    assert new_contents == original_contents



import pytest
from mock import AsyncMock, MagicMock

from opentrons.drivers.asyncio.communication.serial_connection import (
    AsyncResponseSerialConnection,
)
from opentrons.drivers.vacuum_module import types
from opentrons.drivers.vacuum_module.driver import (
    VacuumModuleDriver,
)


@pytest.fixture
def connection() -> AsyncMock:
    return AsyncMock(spec=AsyncResponseSerialConnection)


@pytest.fixture
def subject(connection: AsyncMock) -> VacuumModuleDriver:
    connection.send_command.return_value = ""
    connection._serial = MagicMock()
    return VacuumModuleDriver(connection)


async def test_get_device_info(
    subject: VacuumModuleDriver, connection: AsyncMock
) -> None:
    """It should send a get device info command"""
    connection.send_command.side_effect = [
        "M115 FW:0.0.1 HW:Opentrons-vacuum-module-nff SerialNo:VMA120230605001",
        "M114 R:0",
    ]
    response = await subject.get_device_info()
    assert response == types.VacuumModuleInfo(
        fw="0.0.1",
        hw=types.HardwareRevision.NFF,
        sn="VMA120230605001",
        rr=0,
    )

    device_info = types.GCODE.GET_DEVICE_INFO.build_command()
    reset_reason = types.GCODE.GET_RESET_REASON.build_command()
    connection.send_command.assert_any_call(device_info)
    connection.send_command.assert_called_with(reset_reason)
    connection.reset_mock()

    # Test invalid response
    connection.send_command.side_effect = [
        "M115 FW:0.0.1 SerialNo:VMA120230605001",
        "M114 R:0",
    ]

    # This should raise ValueError
    with pytest.raises(ValueError):
        response = await subject.get_device_info()

    device_info = types.GCODE.GET_DEVICE_INFO.build_command()
    connection.send_command.assert_any_call(device_info)
    # M115 response is invalid, so we dont send M114.
    connection.send_command.assert_called_once()


async def test_set_serial_number(
    subject: VacuumModuleDriver, connection: AsyncMock
) -> None:
    """It should send a set serial number command"""
    connection.send_command.return_value = "M996"

    serial_number = "VMA1020250119001"
    await subject.set_serial_number(serial_number)

    set_serial_number = types.GCODE.SET_SERIAL_NUMBER.build_command().add_element(
        serial_number
    )
    connection.send_command.assert_any_call(set_serial_number)
    connection.reset_mock()

    # Test invalid response
    connection.send_command.return_value = "M9nn"
    with pytest.raises(ValueError):
        await subject.set_serial_number(serial_number)

    set_serial_number = types.GCODE.SET_SERIAL_NUMBER.build_command().add_element(
        serial_number
    )
    connection.send_command.assert_any_call(set_serial_number)
    connection.reset_mock()

    # Test invalid serial number
    with pytest.raises(ValueError):
        await subject.set_serial_number("invalid")

    connection.send_command.assert_not_called()
    connection.reset_mock()


async def test_set_led(subject: VacuumModuleDriver, connection: AsyncMock) -> None:
    """It should send a set led command"""
    connection.send_command.return_value = "M200"
    await subject.set_led(1, types.LEDColor.RED)

    set_led = types.GCODE.SET_LED.build_command().add_float("P", 1).add_int("C", 1)
    connection.send_command.assert_any_call(set_led)
    connection.reset_mock()

    # test setting only external leds
    await subject.set_led(1, types.LEDColor.RED, external=True)

    set_led = (
        types.GCODE.SET_LED.build_command()
        .add_float("P", 1)
        .add_int("C", 1)
        .add_int("K", 1)
    )
    connection.send_command.assert_any_call(set_led)
    connection.reset_mock()


async def test_set_vacuum_state(
    subject: VacuumModuleDriver, connection: AsyncMock
) -> None:
    """It should send a set pressure command"""
    connection.send_command.return_value = "M120"

    await subject.set_vacuum_state(True, -600)

    set_pressure = (
        types.GCODE.SET_PRESSURE_STATE.build_command()
        .add_int("S", 1)
        .add_float("P", -600)
    )

    connection.send_command.assert_any_call(set_pressure)
    connection.reset_mock()

    # With duration and rate
    connection.send_command.return_value = "M120"

    await subject.set_vacuum_state(True, -600, 500, rate=-10)

    set_pressure = (
        types.GCODE.SET_PRESSURE_STATE.build_command()
        .add_int("S", 1)
        .add_float("P", -600)
        .add_int("D", 500)
        .add_float("R", -10)
    )

    connection.send_command.assert_any_call(set_pressure)
    connection.reset_mock()

    # With vent_after
    connection.send_command.return_value = "M120"

    await subject.set_vacuum_state(True, -600, 500, vent_after=True)

    set_pressure = (
        types.GCODE.SET_PRESSURE_STATE.build_command()
        .add_int("S", 1)
        .add_float("P", -600)
        .add_int("D", 500)
        .add_int("V", 1)
    )

    connection.send_command.assert_any_call(set_pressure)
    connection.reset_mock()

    # vacuum off
    connection.send_command.return_value = "M120"

    await subject.set_vacuum_state(False)

    set_pressure = types.GCODE.SET_PRESSURE_STATE.build_command().add_int("S", 0)

    connection.send_command.assert_any_call(set_pressure)
    connection.reset_mock()


async def test_get_vacuum_state(
    subject: VacuumModuleDriver, connection: AsyncMock
) -> None:
    """It should send a get pressure command"""
    connection.send_command.return_value = (
        "M121 T:400.0 C:988.0 A:989.3 B:988.6 H:992.5 E:1 V:1"
    )

    pressure_state = await subject.get_vacuum_state()

    get_pressure = types.GCODE.GET_PRESSURE_STATE.build_command()
    connection.send_command.assert_any_call(get_pressure)
    connection.reset_mock()

    assert pressure_state == types.PressureState(
        400, 988, 989.3, 988.6, 992.5, True, types.VentState.CLOSED
    )

    # test idle
    connection.send_command.return_value = (
        "M121 T:0.0 C:0.0 A:989.3 B:988.6 H:992.5 E:0 V:0"
    )

    pressure_state = await subject.get_vacuum_state()

    get_pressure = types.GCODE.GET_PRESSURE_STATE.build_command()
    connection.send_command.assert_any_call(get_pressure)
    connection.reset_mock()

    assert pressure_state == types.PressureState(
        0, 0, 989.3, 988.6, 992.5, False, types.VentState.OPENED
    )


async def test_set_pump_state(
    subject: VacuumModuleDriver, connection: AsyncMock
) -> None:
    """It should send a set pump command"""
    connection.send_command.return_value = "M122"

    await subject.set_pump_state(True, 1600)

    set_pump = (
        types.GCODE.SET_PUMP_STATE.build_command().add_int("S", 1).add_float("R", 1600)
    )

    connection.send_command.assert_any_call(set_pump)
    connection.reset_mock()

    # Set Duty cycle directly instead of rpm
    await subject.set_pump_state(True, duty_cycle=30)

    set_pump = (
        types.GCODE.SET_PUMP_STATE.build_command().add_int("S", 1).add_int("D", 30)
    )
    connection.send_command.assert_any_call(set_pump)
    connection.reset_mock()

    # Turn off pump
    await subject.set_pump_state(False)

    set_pump = types.GCODE.SET_PUMP_STATE.build_command().add_int("S", 0)
    connection.send_command.assert_any_call(set_pump)
    connection.reset_mock()


async def test_get_pump_state(
    subject: VacuumModuleDriver, connection: AsyncMock
) -> None:
    """It should send a get pump command"""
    connection.send_command.return_value = "M123 T:1600.0 R:988.0 A:0 D:30 E:1 M:0"

    pump_state = await subject.get_pump_state()

    get_pump = types.GCODE.GET_PUMP_STATE.build_command()
    connection.send_command.assert_any_call(get_pump)
    connection.reset_mock()

    assert pump_state == types.PumpState(1600, 988, 0, 30, True, False)

    # manual mode on
    connection.send_command.return_value = "M123 T:0.0 R:1088.0 A:45 D:30 E:1 M:1"

    pump_state = await subject.get_pump_state()

    get_pump = types.GCODE.GET_PUMP_STATE.build_command()
    connection.send_command.assert_any_call(get_pump)
    connection.reset_mock()

    assert pump_state == types.PumpState(0, 1088, 45, 30, True, True)

    # pump off
    connection.send_command.return_value = "M123 T:0.0 R:0.0 A:0 D:0 E:0 M:1"

    pump_state = await subject.get_pump_state()

    get_pump = types.GCODE.GET_PUMP_STATE.build_command()
    connection.send_command.assert_any_call(get_pump)
    connection.reset_mock()

    assert pump_state == types.PumpState(0, 0, 0, 0, False, True)



from typing import Optional

from .abstract import AbstractVacuumModuleDriver
from .types import (
    HardwareRevision,
    LEDColor,
    LEDPattern,
    PressureState,
    PumpState,
    VacuumModuleInfo,
    VentState,
)
from opentrons.util.async_helpers import ensure_yield


class SimulatingDriver(AbstractVacuumModuleDriver):
    def __init__(self, serial_number: Optional[str] = None) -> None:
        self._serial_number = serial_number or "dummySerialFS"
        self.vent_state = VentState.OPENED
        self.vacuum_on = False
        self.pump_enabled = False
        self.pressure_sensor_enabled = False
        self.target_pressure = 0.0
        self.current_pressure = 0.0
        self.target_rpm = 0
        self.current_rpm = 0

    @ensure_yield
    async def connect(self) -> None:
        pass

    @ensure_yield
    async def disconnect(self) -> None:
        pass

    @ensure_yield
    async def is_connected(self) -> bool:
        return True

    def reset_serial_buffers(self) -> None:
        pass

    async def get_device_info(self) -> VacuumModuleInfo:
        return VacuumModuleInfo(
            fw="vacuum-fw", hw=HardwareRevision.NFF, sn=self._serial_number
        )

    async def enter_programming_mode(self) -> None:
        pass

    async def set_serial_number(self, sn: str) -> None:
        self._serial_number = sn

    async def set_led(
        self,
        power: float,
        color: Optional[LEDColor] = None,
        external: Optional[bool] = None,
        pattern: Optional[LEDPattern] = None,
        duration: Optional[int] = None,  # Default firmware duration is 500ms
        reps: Optional[int] = None,  # Default firmware reps is 0
    ) -> None:
        pass

    async def enable_pump(self) -> None:
        self.pump_enabled = True

    async def disable_pump(self) -> None:
        self.pump_enabled = False

    async def get_pump_motor_register(self) -> None:
        """Get the register value of the pump motor driver."""
        pass

    async def get_pressure_sensor_register(self) -> None:
        """Get the register value of the pressure sensor driver."""
        pass

    # TODO: update the pressure arg with the units when we find out which unit
    async def set_vacuum_state(
        self,
        enable_vacuum: bool,
        guage_pressure_mbar: Optional[float] = None,
        duration: Optional[int] = None,
        rate: Optional[float] = None,
        vent_after: Optional[bool] = None,
    ) -> None:
        """Engage or release the vacuum until a desired internal pressure is reached."""
        self.vacuum_on = enable_vacuum
        self.target_pressure = guage_pressure_mbar or self.target_pressure

    async def get_vacuum_state(self) -> PressureState:
        """Get the pressure state."""
        return PressureState(
            self.target_pressure,
            self.current_pressure,
            0,
            0,
            0,
            self.vacuum_on,
            self.vent_state,
        )

    async def set_pump_state(
        self,
        start_pump: bool,
        target_rpm: Optional[int] = None,
        duty_cycle: Optional[int] = None,
    ) -> None:
        """Start or the stop the pump at a given rpm or duty cycle."""
        self.pump_enabled = start_pump
        self.target_rpm = target_rpm or self.target_rpm

    async def get_pump_state(self) -> PumpState:
        """Get the pump state."""
        return PumpState(0, 0, 0, 0, False, False)

    async def set_vent_state(self, state: bool) -> None:
        """Opens/Closes the vent, which release the vacuum in the module chamber."""
        self.vent_state = VentState(open)



"""Tests for `slot_standardization`."""

import pytest

from opentrons_shared_data.robot.types import RobotType

from opentrons.protocol_engine import (
    OFF_DECK_LOCATION,
    CommandIntent,
    DeckSlotLocation,
    LabwareMovementStrategy,
    LabwareOffsetVector,
    LoadableLabwareLocation,
    ModuleLocation,
    ModuleModel,
    OnLabwareLocation,
    commands,
)
from opentrons.protocol_engine import (
    slot_standardization as subject,
)
from opentrons.types import DeckSlotName


@pytest.mark.parametrize(
    ("original_location", "robot_type", "expected_location"),
    [
        # DeckSlotLocations should have their slotName standardized.
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
            "OT-2 Standard",
            DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        ),
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
            "OT-3 Standard",
            DeckSlotLocation(slotName=DeckSlotName.SLOT_C2),
        ),
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_C2),
            "OT-2 Standard",
            DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        ),
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_C2),
            "OT-3 Standard",
            DeckSlotLocation(slotName=DeckSlotName.SLOT_C2),
        ),
        # ModuleLocations, OnLabwareLocations, and OFF_DECK_LOCATIONs should be left alone.
        (
            ModuleLocation(moduleId="module-id"),
            "OT-3 Standard",
            ModuleLocation(moduleId="module-id"),
        ),
        (
            OnLabwareLocation(labwareId="labware-id"),
            "OT-3 Standard",
            OnLabwareLocation(labwareId="labware-id"),
        ),
        (
            OFF_DECK_LOCATION,
            "OT-3 Standard",
            OFF_DECK_LOCATION,
        ),
    ],
)
def test_standardize_load_labware_command(
    original_location: LoadableLabwareLocation,
    robot_type: RobotType,
    expected_location: LoadableLabwareLocation,
) -> None:
    """It should convert deck slots in `LoadLabwareCreate`s."""
    original = commands.LoadLabwareCreate(
        intent=CommandIntent.SETUP,
        key="key",
        params=commands.LoadLabwareParams(
            location=original_location,
            loadName="loadName",
            namespace="namespace",
            version=123,
            labwareId="labwareId",
            displayName="displayName",
        ),
    )
    expected = commands.LoadLabwareCreate(
        intent=CommandIntent.SETUP,
        key="key",
        params=commands.LoadLabwareParams(
            location=expected_location,
            loadName="loadName",
            namespace="namespace",
            version=123,
            labwareId="labwareId",
            displayName="displayName",
        ),
    )
    assert subject.standardize_command(original, robot_type) == expected


@pytest.mark.parametrize(
    ("original_location", "robot_type", "expected_location"),
    [
        # DeckSlotLocations should have their slotName standardized.
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
            "OT-2 Standard",
            DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        ),
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
            "OT-3 Standard",
            DeckSlotLocation(slotName=DeckSlotName.SLOT_C2),
        ),
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_C2),
            "OT-2 Standard",
            DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        ),
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_C2),
            "OT-3 Standard",
            DeckSlotLocation(slotName=DeckSlotName.SLOT_C2),
        ),
        # ModuleLocations and OFF_DECK_LOCATIONs should be left alone.
        (
            ModuleLocation(moduleId="module-id"),
            "OT-3 Standard",
            ModuleLocation(moduleId="module-id"),
        ),
        (
            OnLabwareLocation(labwareId="labware-id"),
            "OT-3 Standard",
            OnLabwareLocation(labwareId="labware-id"),
        ),
        (
            OFF_DECK_LOCATION,
            "OT-3 Standard",
            OFF_DECK_LOCATION,
        ),
    ],
)
def test_standardize_move_labware_command(
    original_location: LoadableLabwareLocation,
    robot_type: RobotType,
    expected_location: LoadableLabwareLocation,
) -> None:
    """It should convert deck slots in `MoveLabwareCreate`s."""
    original = commands.MoveLabwareCreate(
        intent=CommandIntent.SETUP,
        key="key",
        params=commands.MoveLabwareParams(
            newLocation=original_location,
            labwareId="labwareId",
            strategy=LabwareMovementStrategy.USING_GRIPPER,
            pickUpOffset=LabwareOffsetVector(x=1, y=2, z=3),
            dropOffset=LabwareOffsetVector(x=4, y=5, z=6),
        ),
    )
    expected = commands.MoveLabwareCreate(
        intent=CommandIntent.SETUP,
        key="key",
        params=commands.MoveLabwareParams(
            newLocation=expected_location,
            labwareId="labwareId",
            strategy=LabwareMovementStrategy.USING_GRIPPER,
            pickUpOffset=LabwareOffsetVector(x=1, y=2, z=3),
            dropOffset=LabwareOffsetVector(x=4, y=5, z=6),
        ),
    )
    assert subject.standardize_command(original, robot_type) == expected


@pytest.mark.parametrize(
    ("original_location", "robot_type", "expected_location"),
    [
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
            "OT-2 Standard",
            DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        ),
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
            "OT-3 Standard",
            DeckSlotLocation(slotName=DeckSlotName.SLOT_C2),
        ),
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_C2),
            "OT-2 Standard",
            DeckSlotLocation(slotName=DeckSlotName.SLOT_5),
        ),
        (
            DeckSlotLocation(slotName=DeckSlotName.SLOT_C2),
            "OT-3 Standard",
            DeckSlotLocation(slotName=DeckSlotName.SLOT_C2),
        ),
    ],
)
def test_standardize_load_module_command(
    original_location: DeckSlotLocation,
    robot_type: RobotType,
    expected_location: DeckSlotLocation,
) -> None:
    """It should convert deck slots in `LoadModuleCreate`s."""
    original = commands.LoadModuleCreate(
        intent=CommandIntent.SETUP,
        key="key",
        params=commands.LoadModuleParams(
            location=original_location,
            model=ModuleModel.MAGNETIC_MODULE_V1,
            moduleId="moduleId",
        ),
    )
    expected = commands.LoadModuleCreate(
        intent=CommandIntent.SETUP,
        key="key",
        params=commands.LoadModuleParams(
            location=expected_location,
            model=ModuleModel.MAGNETIC_MODULE_V1,
            moduleId="moduleId",
        ),
    )
    assert subject.standardize_command(original, robot_type) == expected


@pytest.mark.parametrize("robot_type", ["OT-2 Standard", "OT-3 Standard"])
def test_standardize_other_commands(robot_type: RobotType) -> None:
    """If a`CommandCreate` contains no deck slot, it should be passed through unchanged."""
    original = commands.HomeCreate(params=commands.HomeParams())
    assert subject.standardize_command(original, robot_type) == original



"""Liquid view tests.

DEPRECATED: Testing LiquidView independently of LiquidStore is no longer helpful.
Try to add new tests to test_liquid_state.py, where they can be tested together,
treating LiquidState as a private implementation detail.
"""

import pytest

from opentrons.protocol_engine import Liquid
from opentrons.protocol_engine.errors import InvalidLiquidError, LiquidDoesNotExistError
from opentrons.protocol_engine.state.liquids import LiquidState, LiquidView


@pytest.fixture
def subject() -> LiquidView:
    """Get a liquid view test subject."""
    state = LiquidState(
        liquids_by_id={
            "water-id": Liquid(
                id="water-id", displayName="water", description="water desc"
            )
        }
    )

    return LiquidView(state)


def test_get_all(subject: LiquidView) -> None:
    """Should return a list of liquids."""
    assert subject.get_all() == [
        Liquid(id="water-id", displayName="water", description="water desc")
    ]


def test_has_liquid(subject: LiquidView) -> None:
    """Should return true if an item exists in the liquids list."""
    assert subject.validate_liquid_id("water-id") == "water-id"

    with pytest.raises(LiquidDoesNotExistError):
        subject.validate_liquid_id("no-id")


def test_validate_liquid_prevents_empty(subject: LiquidView) -> None:
    """It should not allow loading a liquid with the special id EMPTY."""
    with pytest.raises(InvalidLiquidError):
        subject.validate_liquid_allowed(
            Liquid(id="EMPTY", displayName="empty", description="nothing")
        )


def test_validate_liquid_allows_non_empty(subject: LiquidView) -> None:
    """It should allow a valid liquid."""
    valid_liquid = Liquid(
        id="some-id",
        displayName="some-display-name",
        description="some-description",
        displayColor=None,
    )
    assert subject.validate_liquid_allowed(valid_liquid) == valid_liquid



from typing import AsyncGenerator, Iterator

import pytest

from opentrons import _find_smoothie_file
from opentrons.config.robot_configs import build_config
from opentrons.config.types import RobotConfig
from opentrons.hardware_control import Controller
from opentrons.hardware_control.emulation.settings import Settings
from opentrons.types import Mount


@pytest.fixture
async def subject(
    emulation_app: Iterator[None],
    emulator_settings: Settings,
) -> AsyncGenerator[Controller, None]:
    conf = build_config({})
    port = f"socket://127.0.0.1:{emulator_settings.smoothie.port}"
    assert isinstance(conf, RobotConfig)
    hc = await Controller.build(config=conf)
    await hc.connect(port=port)
    try:
        yield hc
    finally:
        await hc.clean_up()
        await hc._smoothie_driver.disconnect()


async def test_get_fw_version(subject: Controller) -> None:
    """It should be set."""
    _, fw_version = _find_smoothie_file()
    assert subject._cached_fw_version == fw_version


async def test_get_attached_instruments(subject: Controller) -> None:
    """It should get the attached instruments."""
    instruments = await subject.get_attached_instruments({})
    assert instruments[Mount.RIGHT]["id"] == "P20SV202020070101"
    right_config = instruments[Mount.RIGHT]["config"]
    assert right_config is not None
    assert right_config.display_name == "P20 Single-Channel GEN2"
    assert instruments[Mount.LEFT]["id"] == "P3HMV202020041605"
    left_config = instruments[Mount.LEFT]["config"]
    assert left_config is not None
    assert left_config.display_name == "P20 8-Channel GEN2"


async def test_move(subject: Controller) -> None:
    """It should move."""
    new_position = {"X": 1.0, "Z": 2.0, "Y": 3.0, "A": 4.0, "B": 5.0, "C": 6.0}

    await subject.move(target_position=new_position)

    updated_position = await subject.update_position()

    assert updated_position == {"X": 1, "Z": 2, "Y": 3, "A": 4, "B": 5, "C": 6}



"""Test for Calibration Set Up Position Implementation."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from decoy import Decoy

from opentrons.hardware_control.types import Axis, CriticalPoint
from opentrons.protocol_engine.commands.calibration.move_to_maintenance_position import (
    MaintenancePosition,
    MoveToMaintenancePositionImplementation,
    MoveToMaintenancePositionParams,
    MoveToMaintenancePositionResult,
)
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.state.state import StateView
from opentrons.types import Mount, MountType, Point

if TYPE_CHECKING:
    from opentrons.hardware_control.ot3api import OT3API


@pytest.fixture
def subject(
    ot3_hardware_api: OT3API, state_view: StateView
) -> MoveToMaintenancePositionImplementation:
    """Returns the subject under test."""
    return MoveToMaintenancePositionImplementation(
        state_view=state_view, hardware_api=ot3_hardware_api
    )


@pytest.mark.ot3_only
@pytest.mark.parametrize("mount_type", [MountType.LEFT, MountType.RIGHT])
async def test_calibration_move_to_location_implementation_for_attach_instrument(
    decoy: Decoy,
    subject: MoveToMaintenancePositionImplementation,
    state_view: StateView,
    ot3_hardware_api: OT3API,
    mount_type: MountType,
) -> None:
    """Command should get a move to target location and critical point and should verify move_to call."""
    params = MoveToMaintenancePositionParams(
        mount=mount_type, maintenancePosition=MaintenancePosition.ATTACH_INSTRUMENT
    )

    decoy.when(
        await ot3_hardware_api.gantry_position(
            Mount.LEFT, critical_point=CriticalPoint.MOUNT
        )
    ).then_return(Point(x=1, y=2, z=250))

    decoy.when(ot3_hardware_api.get_instrument_max_height(Mount.LEFT)).then_return(300)

    result = await subject.execute(params=params)
    assert result == SuccessData(
        public=MoveToMaintenancePositionResult(),
    )

    hw_mount = mount_type.to_hw_mount()
    decoy.verify(
        await ot3_hardware_api.prepare_for_mount_movement(Mount.LEFT),
        await ot3_hardware_api.retract(Mount.LEFT),
        await ot3_hardware_api.move_to(
            mount=Mount.LEFT,
            abs_position=Point(x=0, y=110, z=250),
            critical_point=CriticalPoint.MOUNT,
        ),
        await ot3_hardware_api.prepare_for_mount_movement(hw_mount),
        await ot3_hardware_api.move_axes(
            position={Axis.by_mount(hw_mount): 400},
        ),
        await ot3_hardware_api.disengage_axes(
            [Axis.by_mount(hw_mount)],
        ),
    )


@pytest.mark.ot3_only
@pytest.mark.parametrize("mount_type", [MountType.LEFT, MountType.RIGHT])
async def test_calibration_move_to_location_implementation_for_attach_plate(
    decoy: Decoy,
    subject: MoveToMaintenancePositionImplementation,
    state_view: StateView,
    ot3_hardware_api: OT3API,
    mount_type: MountType,
) -> None:
    """Command should get a move to target location and critical point and should verify move_to call."""
    params = MoveToMaintenancePositionParams(
        mount=mount_type, maintenancePosition=MaintenancePosition.ATTACH_PLATE
    )

    decoy.when(
        await ot3_hardware_api.gantry_position(
            Mount.LEFT, critical_point=CriticalPoint.MOUNT
        )
    ).then_return(Point(x=1, y=2, z=250))

    decoy.when(ot3_hardware_api.get_instrument_max_height(Mount.LEFT)).then_return(300)

    result = await subject.execute(params=params)
    assert result == SuccessData(
        public=MoveToMaintenancePositionResult(),
    )

    decoy.verify(
        await ot3_hardware_api.prepare_for_mount_movement(Mount.LEFT),
        await ot3_hardware_api.retract(Mount.LEFT),
        await ot3_hardware_api.move_to(
            mount=Mount.LEFT,
            abs_position=Point(x=0, y=110, z=250),
            critical_point=CriticalPoint.MOUNT,
        ),
        await ot3_hardware_api.move_axes(
            position={
                Axis.Z_L: 90,
            }
        ),
        await ot3_hardware_api.disengage_axes(
            [Axis.Z_L],
        ),
        await ot3_hardware_api.move_axes(
            position={
                Axis.Z_R: 105,
            }
        ),
        await ot3_hardware_api.disengage_axes(
            [Axis.Z_R],
        ),
    )


@pytest.mark.ot3_only
async def test_calibration_move_to_location_implementation_for_gripper(
    decoy: Decoy,
    subject: MoveToMaintenancePositionImplementation,
    state_view: StateView,
    ot3_hardware_api: OT3API,
) -> None:
    """Command should get a move to target location and critical point and should verify move_to call."""
    params = MoveToMaintenancePositionParams(
        mount=MountType.EXTENSION,
        maintenancePosition=MaintenancePosition.ATTACH_INSTRUMENT,
    )

    decoy.when(
        await ot3_hardware_api.gantry_position(
            Mount.LEFT, critical_point=CriticalPoint.MOUNT
        )
    ).then_return(Point(x=1, y=2, z=250))
    decoy.when(ot3_hardware_api.get_instrument_max_height(Mount.LEFT)).then_return(300)

    result = await subject.execute(params=params)
    assert result == SuccessData(
        public=MoveToMaintenancePositionResult(),
    )

    decoy.verify(
        await ot3_hardware_api.prepare_for_mount_movement(Mount.LEFT),
        await ot3_hardware_api.retract(Mount.LEFT),
        await ot3_hardware_api.move_to(
            mount=Mount.LEFT,
            abs_position=Point(x=0, y=110, z=250),
            critical_point=CriticalPoint.MOUNT,
        ),
    )

    decoy.verify(
        await ot3_hardware_api.move_axes(
            position={Axis.Z_G: 400},
        ),
        times=0,
    )
    decoy.verify(
        await ot3_hardware_api.disengage_axes(
            [Axis.Z_G],
        ),
        times=0,
    )
